from typing import List, Optional, Union

from foundry.game.gfx.GraphicsSet import GraphicsSet, GraphicsSetProtocol
from foundry.game.gfx.objects.Jump import Jump
from foundry.game.gfx.objects.LevelObject import (
    SCREEN_HEIGHT,
    SCREEN_WIDTH,
    LevelObject,
)
from foundry.game.gfx.Palette import PaletteGroup


class LevelObjectFactory:
    object_set: int
    graphic_set: int
    palette_group_index: int

    graphics_set: Optional[GraphicsSetProtocol] = None
    palette_group: list = []

    def __init__(
        self,
        object_set: int,
        graphic_set: Union[int, GraphicsSetProtocol],
        palette_group_index: int,
        objects_ref: List[LevelObject],
        vertical_level: bool,
        size_minimal: bool = False,
    ):
        self.set_object_set(object_set)
        if isinstance(graphic_set, int):
            self.set_graphic_set(graphic_set)
        else:
            self.graphic_set = 0
            self.graphics_set = graphic_set
        self.set_palette_group_index(palette_group_index)
        self.objects_ref = objects_ref
        self.vertical_level = vertical_level

        self.size_minimal = size_minimal

    def set_object_set(self, object_set: int):
        self.object_set = object_set

    def set_graphic_set(self, graphic_set: int):
        self.graphic_set = graphic_set
        self.graphics_set = GraphicsSet.from_tileset(self.graphic_set)

    def set_palette_group_index(self, palette_group_index: int):
        self.palette_group_index = palette_group_index
        self.palette_group = PaletteGroup.from_tileset(self.object_set, self.palette_group_index)

    def from_data(self, data: bytearray, index: int):
        if Jump.is_jump(data):
            return Jump(data)

        assert self.graphics_set is not None

        # todo get rid of index by fixing ground map
        return LevelObject(
            data,
            self.object_set,
            self.palette_group,
            self.graphics_set,
            self.objects_ref,
            self.vertical_level,
            index,
            size_minimal=self.size_minimal,
        )

    def from_properties(
        self,
        domain: int,
        object_index: int,
        x: int,
        y: int,
        length: Optional[int],
        index: int,
    ):
        assert isinstance(domain, int)
        assert isinstance(object_index, int)
        assert isinstance(x, int)
        assert isinstance(y, int)
        assert isinstance(index, int)
        if self.vertical_level:
            offset = y // SCREEN_HEIGHT
            y %= SCREEN_HEIGHT

            x += offset * SCREEN_WIDTH

        data = bytearray(3)

        data[0] = domain << 5 | y
        data[1] = x
        data[2] = object_index

        if length is not None:
            data.append(length)

        obj = self.from_data(data, index)

        return obj



from foundry.core.warnings.Warning import PydanticWarning, Warning
from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.game.gfx.objects.ObjectLike import ObjectLike


class ExtendToGroundWarning(Warning):
    """
    A warning to ensure that an object does not extend to the ground.
    """

    def check_object(self, obj: LevelObject, *args, **kwargs) -> bool:
        """
        Determines if the object should emit a warning for extending to the ground.

        Parameters
        ----------
        obj : LevelObject
            The object to check.

        Returns
        -------
        bool
            If the object should emit a warning.
        """
        return obj.position.y + obj.rendered_height == 27

    def get_message(self, obj: ObjectLike) -> str:
        return f"{obj} extends until the level bottom. This can crash the game."


class PydanticExtendToGroundWarning(PydanticWarning):
    """
    A JSON model of a warning that checks that an object does not extend to the ground.
    """



import json

from qt_material import build_stylesheet

from foundry import default_settings_path

RESIZE_LEFT_CLICK = "LMB"
RESIZE_RIGHT_CLICK = "RMB"


def set_style(theme):
    def wrapped(app):
        app.setStyleSheet(build_stylesheet(theme))

    return wrapped


GUI_STYLE = {
    "DARK AMBER": set_style("dark_amber.xml"),
    "DARK BLUE": set_style("dark_blue.xml"),
    "DARK CYAN": set_style("dark_cyan.xml"),
    "DARK GREEN": set_style("dark_lightgreen.xml"),
    "DARK PINK": set_style("dark_pink.xml"),
    "DARK PURPLE": set_style("dark_purple.xml"),
    "DARK RED": set_style("dark_red.xml"),
    "DARK TEAL": set_style("dark_teal.xml"),
    "DARK YELLOW": set_style("dark_yellow.xml"),
    "LIGHT AMBER": set_style("light_amber.xml"),
    "LIGHT BLUE": set_style("light_blue.xml"),
    "LIGHT CYAN": set_style("light_cyan.xml"),
    "LIGHT GREEN": set_style("light_lightgreen.xml"),
    "LIGHT PINK": set_style("light_pink.xml"),
    "LIGHT PURPLE": set_style("light_purple.xml"),
    "LIGHT RED": set_style("light_red.xml"),
    "LIGHT TEAL": set_style("light_teal.xml"),
    "LIGHT YELLOW": set_style("light_yellow.xml"),
}

SETTINGS = dict()
SETTINGS["instaplay_emulator"] = "fceux"
SETTINGS["instaplay_arguments"] = "%f"
SETTINGS["default_powerup"] = 0
SETTINGS["default_power_has_star"] = False
SETTINGS["default_starting_world"] = 0

SETTINGS["resize_mode"] = RESIZE_LEFT_CLICK
SETTINGS["gui_style"] = ""  # initially blank, since we can't call load_stylesheet until the app is started

SETTINGS["draw_mario"] = True
SETTINGS["draw_jumps"] = False
SETTINGS["draw_grid"] = False
SETTINGS["draw_expansion"] = False
SETTINGS["draw_jump_on_objects"] = True
SETTINGS["draw_items_in_blocks"] = True
SETTINGS["draw_invisible_items"] = True
SETTINGS["draw_autoscroll"] = False
SETTINGS["block_transparency"] = True
SETTINGS["object_scroll_enabled"] = False
SETTINGS["object_tooltip_enabled"] = True


def load_settings():
    if not default_settings_path.exists():
        return

    try:
        with open(str(default_settings_path), "r") as settings_file:
            settings_dict = json.loads(settings_file.read())
    except json.JSONDecodeError:
        return

    SETTINGS.update(settings_dict)


def save_settings():
    with open(str(default_settings_path), "w") as settings_file:
        settings_file.write(json.dumps(SETTINGS, indent=4, sort_keys=True))



from typing import Protocol

from attr import attrs

from foundry.core.Position import PositionProtocol


class SpriteProtocol(Protocol):
    position: PositionProtocol
    index: int
    palette_index: int
    horizontal_mirror: bool
    vertical_mirror: bool
    do_not_render: bool


@attrs(slots=True, auto_attribs=True)
class Sprite:
    """
    A representation of a Sprite inside the game.

    Attributes
    ----------
    position: PositionProtocol
        The position of the sprite.
    index: int
        The index into the graphics set of the sprite.
    palette_index: int
        The palette index of the sprite into the PaletteGroupProtocol.
    horizontal_mirror: bool
        If the sprite should be horizontally flipped.
    vertical_mirror: bool
        If the sprite should be vertically flipped.
    do_not_render: bool
        If the sprite should not render.
    """

    position: PositionProtocol
    index: int
    palette_index: int
    horizontal_mirror: bool = False
    vertical_mirror: bool = False
    do_not_render: bool = False



from PySide6.QtCore import QPoint, QPointF, QRectF, QSizeF
from PySide6.QtGui import QBrush, QPainter, QPen, QPolygonF, Qt

from foundry.game.File import ROM
from foundry.game.gfx.drawable.Block import Block
from foundry.game.gfx.objects.LevelObject import GROUND, SCREEN_WIDTH
from foundry.game.level.Level import Level
from foundry.smb3parse.constants import (
    AScroll_HorizontalInitMove,
    AScroll_Movement,
    AScroll_MovementLoop,
    AScroll_MovementLoopStart,
    AScroll_MovementLoopTicks,
    AScroll_MovementRepeat,
    AScroll_VelAccel,
)

HORIZONTAL_SCROLL_0 = 0
HORIZONTAL_SCROLL_1 = 1
UP_RIGHT_DIAG_SCROLL = 2
SPIKE_CEILING_SCROLL = 3
UP_TIL_DOOR_SCROLL = 4
WATER_LEVEL_SCROLL = 5


_ASCROLL_SCREEN_HEIGHT = 12


class AutoScrollDrawer:
    def __init__(self, auto_scroll_row: int, level: Level):
        self.auto_scroll_row = auto_scroll_row
        self.level = level

        self.current_pos = QPointF()
        self.horizontal_speed = 0
        self.vertical_speed = 0

        self.rom = ROM()

        self.pixel_length = 1

        self.acceleration_pen = Qt.NoPen
        self.acceleration_brush = Qt.NoBrush
        self.scroll_pen = Qt.NoPen
        self.scroll_brush = Qt.NoBrush

        self.screen_polygon = QPolygonF()

    def draw(self, painter: QPainter, block_length: int):
        self.pixel_length = block_length / Block.WIDTH

        self.scroll_brush = QBrush(Qt.blue)
        self.scroll_pen = QPen(self.scroll_brush, 2 * self.pixel_length)

        self.acceleration_brush = QBrush(Qt.red)
        self.acceleration_pen = QPen(self.acceleration_brush, 2 * self.pixel_length)

        painter.setPen(self.scroll_pen)
        painter.setBrush(self.scroll_brush)

        auto_scroll_type_index = self.auto_scroll_row >> 4
        auto_scroll_routine_index = self.auto_scroll_row & 0b0001_1111

        if auto_scroll_type_index in [
            SPIKE_CEILING_SCROLL,
            UP_TIL_DOOR_SCROLL,
            WATER_LEVEL_SCROLL,
            UP_RIGHT_DIAG_SCROLL,
        ]:
            # not visualized
            return
        elif auto_scroll_type_index not in [HORIZONTAL_SCROLL_0, HORIZONTAL_SCROLL_1]:
            # illegal value, those appear in the vanilla ROM, though; so error out
            return

        first_movement_command_index = (self.rom.int(AScroll_HorizontalInitMove + auto_scroll_routine_index) + 1) % 256
        last_movement_command_index = (self.rom.int(AScroll_HorizontalInitMove + auto_scroll_routine_index + 1)) % 256

        self.horizontal_speed = 0
        self.vertical_speed = 0

        self.current_pos = self._determine_auto_scroll_start(block_length)

        for movement_command_index in range(first_movement_command_index, last_movement_command_index + 1):

            movement_command = self.rom.int(AScroll_Movement + movement_command_index)
            movement_repeat = self.rom.int(AScroll_MovementRepeat + movement_command_index)

            self._execute_movement_command(painter, movement_command, movement_repeat)

        stop_marker = QRectF(QPoint(0, 0), QSizeF(10, 10) * self.pixel_length)
        stop_marker.moveCenter(self.current_pos)

        painter.setPen(Qt.NoPen)
        painter.drawRect(stop_marker)

        painter.setPen(self.scroll_pen)
        painter.setBrush(self.scroll_brush)

        painter.setOpacity(0.2)
        painter.drawPolygon(self.screen_polygon)

    def _execute_movement_command(self, painter: QPainter, command: int, repeat: int):
        h_updates_per_tick = 4  # got those by reading the auto scroll routine
        v_updates_per_tick = 2

        is_acceleration_command = (command >> 4) == 0

        if is_acceleration_command:
            # set speed
            h_acceleration_index = (command & 0b00001100) >> 2
            v_acceleration_index = command & 0b00000011

            assert h_acceleration_index != 3
            assert v_acceleration_index != 3

            h_acceleration = self.rom.int(AScroll_VelAccel + h_acceleration_index)
            v_acceleration = self.rom.int(AScroll_VelAccel + v_acceleration_index)

            if h_acceleration == 0xFF:
                h_acceleration = -0x01

            if v_acceleration == 0xFF:
                v_acceleration = -0x01

            h_acceleration <<= 4
            v_acceleration <<= 4

            movement_ticks = repeat
            repeat = 1
        else:
            auto_scroll_loop_selector = command >> 4

            loop_start_offset = AScroll_MovementLoopStart - 2 + auto_scroll_loop_selector

            if auto_scroll_loop_selector in [0, 1]:
                # normal movement command
                movement_ticks = self.rom.int(loop_start_offset)

                h_acceleration = 0
                v_acceleration = 0
            else:
                # loop command
                movement_loop_start_index = self.rom.int(loop_start_offset)
                movement_loop_end_index = self.rom.int(loop_start_offset + 1)

                number_of_commands = movement_loop_end_index - movement_loop_start_index

                movement_loop_commands = self.rom.read(
                    AScroll_MovementLoop + movement_loop_start_index, number_of_commands
                )
                movement_loop_repeats = self.rom.read(
                    AScroll_MovementLoopTicks + movement_loop_start_index, number_of_commands
                )

                for _ in range(repeat):
                    for sub_command, sub_repeat in zip(movement_loop_commands, movement_loop_repeats):
                        self._execute_movement_command(painter, sub_command, sub_repeat)

                return

        if is_acceleration_command and (h_acceleration or v_acceleration):
            painter.setPen(self.acceleration_pen)
            painter.setBrush(self.acceleration_brush)
        else:
            painter.setPen(self.scroll_pen)
            painter.setBrush(self.scroll_brush)

        # circle at start of new command
        painter.drawEllipse(self.current_pos, 4 * self.pixel_length, 4 * self.pixel_length)

        self._add_points_for_position(self.current_pos)

        if is_acceleration_command and (h_acceleration or v_acceleration):
            for _ in range(movement_ticks):
                self.horizontal_speed += h_acceleration
                self.vertical_speed += v_acceleration

                old_pos = self.current_pos

                self.current_pos += (
                    QPointF(
                        h_updates_per_tick * self.horizontal_speed / 256, v_updates_per_tick * self.vertical_speed / 256
                    )
                    * self.pixel_length
                )

                painter.drawLine(old_pos, self.current_pos)
                self._add_points_for_position(self.current_pos)
        else:
            old_pos = QPointF(self.current_pos)

            h_movement = h_updates_per_tick * self.horizontal_speed / 256 * movement_ticks * repeat
            v_movement = v_updates_per_tick * self.vertical_speed / 256 * movement_ticks * repeat

            self.current_pos += QPointF(h_movement, v_movement) * self.pixel_length

            painter.drawLine(old_pos, self.current_pos)

            self._add_points_for_line(old_pos, self.current_pos)

    def _add_points_for_position(self, pos: QPointF):
        self.screen_polygon = self.screen_polygon.united(QPolygonF.fromList(self._rect_for_point(pos)))

    def _add_points_for_line(self, start: QPointF, stop: QPointF):
        start_points = self._rect_for_point(start)
        stop_points = self._rect_for_point(stop)

        point_list = []

        if start.y() == stop.y():
            point_list.extend([start_points[0], stop_points[1], stop_points[2], start_points[3]])
        elif start.y() < stop.y():
            point_list.extend(start_points[0:2])
            point_list.extend(stop_points[1:4])
            point_list.append(start_points[3])
        else:
            point_list.append(start_points[0])
            point_list.extend(stop_points[0:3])
            point_list.extend(start_points[2:4])

        self.screen_polygon = self.screen_polygon.united(QPolygonF.fromList(point_list))

    def _rect_for_point(self, pos: QPointF):
        top_right = pos + QPointF(SCREEN_WIDTH // 2, -_ASCROLL_SCREEN_HEIGHT // 2) * self.pixel_length * Block.WIDTH
        bottom_right = pos + QPoint(SCREEN_WIDTH // 2, _ASCROLL_SCREEN_HEIGHT // 2) * self.pixel_length * Block.WIDTH

        top_left = top_right - QPointF(SCREEN_WIDTH, 0) * self.pixel_length * Block.WIDTH
        bottom_left = bottom_right - QPointF(SCREEN_WIDTH, 0) * self.pixel_length * Block.WIDTH

        return top_left, top_right, bottom_right, bottom_left

    def _determine_auto_scroll_start(self, block_length: int) -> QPointF:
        # only support horizontal levels for now
        _, mario_y = self.level.header.mario_position()

        scroll_x, scroll_y = SCREEN_WIDTH // 2, min(mario_y + 2, GROUND - _ASCROLL_SCREEN_HEIGHT // 2)

        return QPointF(scroll_x, scroll_y) * block_length



from typing import Optional

from PySide6.QtCore import QMargins, QSize, Signal, SignalInstance
from PySide6.QtGui import QCloseEvent, QKeyEvent, QMouseEvent, Qt
from PySide6.QtWidgets import (
    QComboBox,
    QDialog,
    QGridLayout,
    QLabel,
    QListWidget,
    QPushButton,
    QScrollArea,
    QScrollBar,
    QTabWidget,
    QToolTip,
    QVBoxLayout,
    QWidget,
)

from foundry.game.gfx.drawable.Block import Block
from foundry.game.level.Level import Level
from foundry.game.level.util import Level as LevelMeta
from foundry.game.level.util import get_world_levels
from foundry.game.level.WorldMap import WorldMap
from foundry.gui.Spinner import Spinner
from foundry.gui.WorldMapView import WorldMapView
from foundry.smb3parse.levels import WORLD_COUNT

WORLD_ITEMS = [
    "World Maps",
    "World 1",
    "World 2",
    "World 3",
    "World 4",
    "World 5",
    "World 6",
    "World 7",
    "World 8",
    "Lost Levels",
]

OBJECT_SET_ITEMS = [
    "0 Overworld",
    "1 Plains",
    "2 Dungeon",
    "3 Hilly",
    "4 Sky",
    "5 Piranha Plant",
    "6 Water",
    "7 Mushroom",
    "8 Pipe",
    "9 Desert",
    "A Ship",
    "B Giant",
    "C Ice",
    "D Cloudy",
    "E Underground",
    "F Spade Bonus",
]


OVERWORLD_MAPS_INDEX = 0
WORLD_1_INDEX = 1


def select_by_world_and_level(world: int, level: int) -> LevelMeta:
    return get_world_levels(world, Level.offsets)[level]


class LevelSelector(QDialog):
    def __init__(self, parent, *, start_level: Optional[tuple[int, int]] = None):
        super(LevelSelector, self).__init__(parent)

        self.setWindowTitle("Level Selector")
        self.setModal(True)

        self.level_name = ""

        self.object_set = 0
        self.object_data_offset = 0x0
        self.enemy_data_offset = 0x0

        self.world_label = QLabel(parent=self, text="World")
        self.world_list = QListWidget(parent=self)
        self.world_list.addItems(WORLD_ITEMS)

        self.world_list.itemDoubleClicked.connect(self.on_ok)
        self.world_list.itemSelectionChanged.connect(self.on_world_click)

        self.level_label = QLabel(parent=self, text="Level")
        self.level_list = QListWidget(parent=self)

        self.level_list.itemDoubleClicked.connect(self.on_ok)
        self.level_list.itemSelectionChanged.connect(self.on_level_click)

        self.enemy_data_label = QLabel(parent=self, text="Enemy Data")
        self.enemy_data_spinner = Spinner(parent=self)

        self.object_data_label = QLabel(parent=self, text="Object Data")
        self.object_data_spinner = Spinner(self)

        self.object_set_label = QLabel(parent=self, text="Object Set")
        self.object_set_dropdown = QComboBox(self)
        self.object_set_dropdown.addItems(OBJECT_SET_ITEMS)

        self.button_ok = QPushButton("Ok", self)
        self.button_ok.clicked.connect(self.on_ok)
        self.button_ok.setFocus()

        self.button_cancel = QPushButton("Cancel", self)
        self.button_cancel.clicked.connect(self.close)

        stock_level_widget = QWidget()
        stock_level_layout = QGridLayout(stock_level_widget)

        stock_level_layout.addWidget(self.world_label, 0, 0)
        stock_level_layout.addWidget(self.level_label, 0, 1)

        stock_level_layout.addWidget(self.world_list, 1, 0)
        stock_level_layout.addWidget(self.level_list, 1, 1)

        self.source_selector = QTabWidget()
        self.source_selector.addTab(stock_level_widget, "Stock Levels")

        for world_number in range(WORLD_COUNT - 1):
            world_number += 1

            world_map_select = WorldMapLevelSelect(world_number)
            world_map_select.level_selected.connect(self._on_level_selected_via_world_map)
            world_map_select.accepted.connect(self.on_ok)

            self.source_selector.addTab(world_map_select, f"W{world_number}")

        data_layout = QGridLayout()

        data_layout.addWidget(self.enemy_data_label, 0, 0)
        data_layout.addWidget(self.object_data_label, 0, 1)
        data_layout.addWidget(self.enemy_data_spinner, 1, 0)
        data_layout.addWidget(self.object_data_spinner, 1, 1)

        data_layout.addWidget(self.object_set_label, 2, 0)
        data_layout.addWidget(self.object_set_dropdown, 2, 1)

        data_layout.addWidget(self.button_ok, 3, 0)
        data_layout.addWidget(self.button_cancel, 3, 1)

        main_layout = QVBoxLayout()
        main_layout.addWidget(self.source_selector)
        main_layout.addLayout(data_layout)

        self.setLayout(main_layout)

        if start_level is None:
            self.world_list.setCurrentRow(1)  # select Level 1-1
            self.on_world_click()
        else:
            self.world_list.setCurrentRow(start_level[0])
            self.level_list.setCurrentRow(start_level[1])

    @property
    def current_level_index(self) -> tuple[int, int]:
        return self.world_list.currentRow(), self.level_list.currentRow()

    def keyPressEvent(self, key_event: QKeyEvent):
        if key_event.key() == Qt.Key_Escape:
            self.reject()

    def on_world_click(self):
        index = self.world_list.currentRow()

        assert index >= 0

        self.level_list.clear()

        self.level_list.addItems([level.display_information.name for level in get_world_levels(index, Level.offsets)])

        if self.level_list.count():
            self.level_list.setCurrentRow(0)

            self.on_level_click()

    def on_level_click(self):
        index = self.level_list.currentRow()

        assert index >= 0

        level_is_overworld = self.world_list.currentRow() == OVERWORLD_MAPS_INDEX

        level = get_world_levels(self.world_list.currentRow(), Level.offsets)[index]
        self.level_name = level.display_information.name

        object_data_for_lvl = level.generator_pointer

        if not level_is_overworld:
            object_data_for_lvl -= Level.HEADER_LENGTH

        if not level_is_overworld:
            enemy_data_for_lvl = level.enemy_pointer
        else:
            enemy_data_for_lvl = 0

        if enemy_data_for_lvl > 0:
            # data in look up table is off by one, since workshop ignores the first byte
            enemy_data_for_lvl -= 1

        self.enemy_data_spinner.setEnabled(not level_is_overworld)

        # if self.world_list.currentRow() >= WORLD_1_INDEX:
        object_set_index = level.tileset
        self.button_ok.setDisabled(level_is_overworld)

        self._fill_in_data(object_set_index, object_data_for_lvl, enemy_data_for_lvl)

    def _fill_in_data(self, object_set: int, layout_address: int, enemy_address: int):
        self.object_set_dropdown.setCurrentIndex(object_set)
        self.object_data_spinner.setValue(layout_address)
        self.enemy_data_spinner.setValue(enemy_address)

    def _on_level_selected_via_world_map(
        self, level_name: str, object_set: int, layout_address: int, enemy_address: int
    ):
        self.level_name = level_name

        self._fill_in_data(object_set, layout_address, enemy_address)

        level_is_overworld = object_set == OVERWORLD_MAPS_INDEX
        self.button_ok.setDisabled(level_is_overworld)
        self.enemy_data_spinner.setDisabled(level_is_overworld)

    def on_ok(self, _=None):
        self.object_set = self.object_set_dropdown.currentIndex()
        self.object_data_offset = self.object_data_spinner.value()
        # skip the first byte, because it seems useless
        self.enemy_data_offset = self.enemy_data_spinner.value() + 1

        self.accept()

    def closeEvent(self, _close_event: QCloseEvent):
        self.reject()


class WorldMapLevelSelect(QScrollArea):
    level_selected: SignalInstance = Signal(str, int, int, int)  # type: ignore
    accepted: SignalInstance = Signal()  # type: ignore

    def __init__(self, world_number: int):
        super(WorldMapLevelSelect, self).__init__()

        self.world = WorldMap(world_number)

        self.world_view = WorldMapView(self, self.world)
        self.world_view.setMouseTracking(True)

        self.setWidget(self.world_view)

        self.setMouseTracking(True)

    def mouseMoveEvent(self, event: QMouseEvent):
        x, y = self.world_view.mapFromParent(event.position().toPoint()).toTuple()

        x //= Block.WIDTH * 2
        y //= Block.HEIGHT * 2

        try:
            level_info = self.world.level_at_position(x, y)

            if level_info is None:
                self.setCursor(Qt.ArrowCursor)

                self.setToolTip("")
                QToolTip.hideText()
            else:
                self.setCursor(Qt.PointingHandCursor)

                level_name = self.world.level_name_at_position(x, y)

                object_set = level_info[0]
                object_set_name = OBJECT_SET_ITEMS[object_set].split(" ", 1)[1]
                layout_address, enemy_address = map(hex, level_info[1:])

                self.setToolTip(
                    f"<b>{level_name}</b><br/>"
                    f"<u>Type:</u> {object_set_name}<br/>"
                    f"<u>Objects:</u> {layout_address}<br/>"
                    f"<u>Enemies:</u> {enemy_address}"
                )
        except ValueError:
            pass

        return super(WorldMapLevelSelect, self).mouseMoveEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent):
        x, y = self.world_view.mapFromParent(event.position().toPoint()).toTuple()

        x //= Block.WIDTH * 2
        y //= Block.HEIGHT * 2

        try:
            level_info = self.world.level_at_position(x, y)
        except ValueError:
            level_info = None

        if level_info is not None:
            self.level_selected.emit(self.world.level_name_at_position(x, y), *level_info)

    def mouseDoubleClickEvent(self, event: QMouseEvent):
        self.mouseReleaseEvent(event)
        self.accepted.emit()

    def sizeHint(self) -> QSize:
        orig_size = super(WorldMapLevelSelect, self).sizeHint()

        scrollbar_width = QScrollBar().sizeHint().width()

        return orig_size.grownBy(QMargins(scrollbar_width, scrollbar_width, 0, 0))



from typing import Optional

from PySide6.QtCore import Signal, SignalInstance
from PySide6.QtWidgets import QHBoxLayout, QWidget

from foundry.game.File import ROM
from foundry.game.gfx.Palette import PaletteGroup
from foundry.game.level.Level import Level
from foundry.gui.PaletteGroupEditor import PaletteGroupEditor
from foundry.gui.PaletteGroupModel import PaletteGroupModel


class PaletteGroupController(QWidget):
    palette_group_changed: SignalInstance = Signal(PaletteGroup, PaletteGroup)  # type: ignore

    def __init__(
        self,
        parent: Optional[QWidget],
        tileset: int = 0,
        bg_offset: int = 0,
        spr_offset: int = 0,
        bg_palette_group: Optional[PaletteGroup] = None,
        spr_palette_group: Optional[PaletteGroup] = None,
    ):
        super().__init__(parent)

        self.model = PaletteGroupModel(
            tileset,
            bg_offset,
            spr_offset,
            bg_palette_group if bg_palette_group is not None else PaletteGroup.as_empty(),
            spr_palette_group if spr_palette_group is not None else PaletteGroup.as_empty(),
        )

        layout = QHBoxLayout()
        layout.setSpacing(0)

        self.bg_widget = PaletteGroupEditor(self, self.model.background_palette_group)
        self.spr_widget = PaletteGroupEditor(self, self.model.sprite_palette_group)

        layout.addWidget(self.bg_widget)
        layout.addWidget(self.spr_widget)

        self.bg_widget.palette_group_changed.connect(lambda *_: self.on_palette_update())
        self.spr_widget.palette_group_changed.connect(lambda *_: self.on_palette_update())

        self.setLayout(layout)

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self.parent}, {self.model.tileset}, "
            + f"{self.model.background_index}, {self.model.sprite_index}, "
            + f"{self.model.background_palette_group}, {self.model.sprite_palette_group})"
        )

    def load_from_level(self, level: Level):
        self.model.tileset = level.object_set_number
        self.model.background_index = level.object_palette_index
        self.model.sprite_index = level.enemy_palette_index
        self.model.restore()
        self._changed = False
        self.silent_update()

    def restore(self):
        self.model.restore()
        self.model.soft_save()
        self._changed = False
        self.silent_update()

    def save(self, rom: Optional[ROM] = None):
        self.model.save(rom)
        self._changed = False

    def silent_update(self):
        self.bg_widget._palette_group = self.model.background_palette_group
        self.bg_widget._update()
        self.spr_widget._palette_group = self.model.sprite_palette_group
        self.spr_widget._update()

    @property
    def changed(self) -> bool:
        return self.model.changed

    @changed.setter
    def changed(self, value: bool):
        self.model.changed = value

    def on_palette_update(self):
        self.model.background_palette_group = self.bg_widget.palette_group
        self.model.sprite_palette_group = self.spr_widget.palette_group
        self.changed = True
        self.model.soft_save()
        self.palette_group_changed.emit(self.model.background_palette_group, self.model.sprite_palette_group)



from typing import Optional

from attr import attrs
from PySide6.QtCore import Signal, SignalInstance
from PySide6.QtGui import QCloseEvent
from PySide6.QtWidgets import QHBoxLayout, QLayout, QStatusBar, QToolBar, QWidget

from foundry import icon
from foundry.core.player_animations import ANIMATION_WIDTH
from foundry.core.player_animations.PlayerAnimation import PlayerAnimation
from foundry.core.player_animations.util import (
    get_animations_palette_index,
    load_player_animation,
)
from foundry.core.Position import Position
from foundry.core.sprites import SPRITE_SIZE
from foundry.core.sprites.Sprite import Sprite, SpriteProtocol
from foundry.core.sprites.SpriteGroup import SpriteGroup, SpriteGroupProtocol
from foundry.game.gfx.Palette import PaletteGroupProtocol
from foundry.gui.CustomChildWindow import CustomChildWindow
from foundry.gui.Spinner import Spinner
from foundry.gui.SpriteViewer import SpriteViewerController as SpriteViewer
from foundry.gui.SpriteViewerWidget import SpriteViewerWidget


@attrs(slots=True, auto_attribs=True)
class PlayerFrameEditorModel:
    animation: PlayerAnimation
    power_up_offsets: list[int]
    palette_group: PaletteGroupProtocol
    is_mario: bool


class PlayerFrameEditorController(CustomChildWindow):
    frames_changed: SignalInstance = Signal(object)  # type: ignore
    offset_changed: SignalInstance = Signal(int)  # type: ignore
    destroyed: SignalInstance = Signal()  # type: ignore

    def __init__(
        self,
        parent: Optional[QWidget],
        animation: PlayerAnimation,
        power_up_offsets: list[int],
        palette_group: PaletteGroupProtocol,
        is_mario: bool = True,
        title: str = "Player Frame Editor",
        zoom: int = 2,
    ):
        super().__init__(parent, title)

        self.model = PlayerFrameEditorModel(animation, power_up_offsets, palette_group, is_mario)
        self.view = PlayerFrameEditorView(self, self.sprite_groups, zoom)
        self.sprite_viewer: Optional[SpriteViewer] = None
        self.sprite_viewer_index = 0
        self.setCentralWidget(self.view)
        self.toolbar = QToolBar(self)

        self.toolbar.setMovable(False)

        self.view.clicked.connect(self._create_sprite_viewer)

        self.zoom_out_action = self.toolbar.addAction(icon("zoom-out.svg"), "Zoom Out")
        self.zoom_in_action = self.toolbar.addAction(icon("zoom-in.svg"), "Zoom In")

        self.page_offset_spinner = Spinner(self, 0xFF)
        self.page_offset_spinner.setValue(animation.offset)

        def change_zoom(offset: int):
            self.view.zoom += offset

        def change_page_offset(offset: int):
            self.page_offset = offset

        self.zoom_out_action.triggered.connect(lambda *_: change_zoom(-1))  # type: ignore
        self.zoom_in_action.triggered.connect(lambda *_: change_zoom(1))  # type: ignore

        self.page_offset_spinner.valueChanged.connect(change_page_offset)  # type: ignore

        self.toolbar.addWidget(self.page_offset_spinner)

        self.addToolBar(self.toolbar)

        self.setStatusBar(QStatusBar(self))
        self.layout().setSizeConstraint(QLayout.SetFixedSize)

    def closeEvent(self, event: QCloseEvent):
        self.toolbar.close()
        if self.sprite_viewer is not None:
            self.sprite_viewer.close()
        self.destroyed.emit()
        super().closeEvent(event)

    @property
    def is_mario(self) -> bool:
        return self.model.is_mario

    @is_mario.setter
    def is_mario(self, is_mario: bool):
        self.model.is_mario = is_mario
        self._update_view_animations()

    @property
    def power_up_offsets(self) -> list[int]:
        return self.model.power_up_offsets

    @power_up_offsets.setter
    def power_up_offsets(self, offsets: list[int]):
        self.model.power_up_offsets = offsets
        self._update_view_animations()

    @property
    def palette_group(self) -> PaletteGroupProtocol:
        return self.model.palette_group

    @palette_group.setter
    def palette_group(self, palette_group: PaletteGroupProtocol):
        self.model.palette_group = palette_group
        self._update_view_animations()

    @property
    def page_offset(self) -> int:
        return self.model.animation.offset

    @page_offset.setter
    def page_offset(self, offset: int):
        self.model.animation.offset = offset
        self.page_offset_spinner.setValue(offset)
        self.offset_changed.emit(offset)
        self._update_view_animations()

    @property
    def frame(self) -> bytearray:
        return self.model.animation.animations

    @frame.setter
    def frame(self, frame: bytearray):
        self.model.animation.animations = frame
        self.frames_changed.emit(frame)
        self._update_view_animations()

    @property
    def sprite_groups(self) -> list[SpriteGroupProtocol]:
        sprite_groups: list[SpriteGroupProtocol] = []

        for idx in range(len(self.model.power_up_offsets)):
            animation = load_player_animation(
                self.model.animation,
                self.model.palette_group,
                self.model.is_mario,
                idx,
                self.model.power_up_offsets,
            )
            sprites: list[SpriteProtocol] = []

            for idx, sprite in enumerate(animation.frames):
                sprites.append(
                    Sprite(
                        Position(
                            (idx % ANIMATION_WIDTH) * SPRITE_SIZE.width,
                            (idx // ANIMATION_WIDTH) * SPRITE_SIZE.height,
                        ),
                        sprite,
                        animation.palette_index,
                        animation.horizontal_flip[idx],
                        do_not_render=sprite == 0xF0,
                    )
                )

            sprite_groups.append(SpriteGroup(Position(0, 0), sprites, animation.graphics_set, animation.palette_group))

        return sprite_groups

    def _update_view_animations(self):
        self.view.sprite_groups = self.sprite_groups
        self.page_offset_spinner.setValue(self.page_offset)
        self.view._update()

    def _create_sprite_viewer(self, positional_index: int, widget_index: int):
        self.sprite_viewer_index = positional_index

        if self.sprite_viewer is None:
            self.sprite_viewer = SpriteViewer(
                self,
                self.sprite_groups[widget_index].graphics_set,
                self.model.palette_group,
                get_animations_palette_index(self.model.is_mario, widget_index),
            )

            def remove_sprite_viewer():
                self.sprite_viewer = None

            def set_sprite(sprite: int):
                sprites = self.frame
                sprites[self.sprite_viewer_index] = sprite * 2
                self.frame = sprites

            self.sprite_viewer.destroyed.connect(remove_sprite_viewer)
            self.sprite_viewer.sprite_selected.connect(set_sprite)
            self.sprite_viewer.show()
        else:
            self.sprite_viewer.graphics_set = self.sprite_groups[widget_index].graphics_set
            self.sprite_viewer.palette_index = get_animations_palette_index(self.model.is_mario, widget_index)


class PlayerFrameEditorView(QWidget):
    clicked: SignalInstance = Signal(int, int)  # type: ignore

    def __init__(
        self,
        parent: Optional[QWidget],
        sprite_groups: list[SpriteGroupProtocol],
        zoom: int = 2,
    ):
        super().__init__(parent)

        self.sprite_groups = sprite_groups
        self._zoom = zoom

        self.layout_ = QHBoxLayout(self)
        self.widgets: list[SpriteViewerWidget] = []

        for idx, sprite_group in enumerate(sprite_groups):
            widget = SpriteViewerWidget(self, sprite_group, zoom=self.zoom)
            widget.clicked.connect(lambda x, y, idx=idx: self.clicked.emit(x + y * ANIMATION_WIDTH, idx))
            self.widgets.append(widget)
            self.layout_.addWidget(widget)

        self.setLayout(self.layout_)

    @property
    def zoom(self) -> int:
        return self._zoom

    @zoom.setter
    def zoom(self, value: int):
        self._zoom = value
        for widget in self.widgets:
            widget.zoom = value

    def _update(self):
        for (widget, sprite_group) in zip(self.widgets, self.sprite_groups):
            widget.sprite_group = sprite_group
            widget.update()



from typing import Optional

from attr import attrs
from PySide6.QtCore import QPoint, QRect, Signal, SignalInstance
from PySide6.QtGui import (
    QBrush,
    QCloseEvent,
    QMouseEvent,
    QPainter,
    QPaintEvent,
    QResizeEvent,
)
from PySide6.QtWidgets import QLayout, QStatusBar, QToolBar, QWidget

from foundry import icon
from foundry.core.Position import Position
from foundry.game.gfx.drawable.Block import Block
from foundry.game.gfx.GraphicsSet import GraphicsSet
from foundry.game.gfx.Palette import PaletteGroup
from foundry.gui.CustomChildWindow import CustomChildWindow
from foundry.gui.PatternViewer import PatternViewerController as PatternViewer


@attrs(slots=True, auto_attribs=True)
class BlockEditorModel:
    tile_square_assembly: bytearray
    block_index: int
    graphics_set: GraphicsSet
    palette_group: PaletteGroup
    palette_index: int


class BlockEditorController(CustomChildWindow):
    tile_square_assembly_changed: SignalInstance = Signal(bytearray)  # type: ignore
    block_index_changed: SignalInstance = Signal(int)  # type: ignore
    graphics_set_changed: SignalInstance = Signal(GraphicsSet)  # type: ignore
    palette_group_changed: SignalInstance = Signal(PaletteGroup)  # type: ignore
    palette_index_changed: SignalInstance = Signal(int)  # type: ignore
    destroyed: SignalInstance = Signal()  # type: ignore

    def __init__(
        self,
        parent: Optional[QWidget],
        tsa_data: bytearray,
        block_index: int,
        graphics_set: GraphicsSet,
        palette_group: PaletteGroup,
        palette_index: int,
    ):
        super().__init__(parent, "Block Editor")

        self.model = BlockEditorModel(tsa_data, block_index, graphics_set, palette_group, palette_index)
        self.view = BlockEditorView(self, tsa_data, block_index, graphics_set, palette_group, palette_index)
        self.pattern_viewer = None
        self._last_x = 0
        self._last_y = 0
        self.setCentralWidget(self.view)
        self.toolbar = QToolBar(self)

        self.view.pattern_selected.connect(lambda values: self._on_pattern_selected(*values))

        self.zoom_out_action = self.toolbar.addAction(icon("zoom-out.svg"), "Zoom Out")
        self.zoom_in_action = self.toolbar.addAction(icon("zoom-in.svg"), "Zoom In")

        def change_zoom(offset: int):
            self.view.zoom += offset

        self.zoom_out_action.triggered.connect(lambda *_: change_zoom(-1))  # type: ignore
        self.zoom_in_action.triggered.connect(lambda *_: change_zoom(1))  # type: ignore

        self.addToolBar(self.toolbar)
        self.setStatusBar(QStatusBar(self))

        self.layout().setSizeConstraint(QLayout.SetFixedSize)

    def closeEvent(self, event: QCloseEvent):
        self.toolbar.close()
        if self.pattern_viewer is not None:
            self.pattern_viewer.close()
        self.destroyed.emit()
        super().closeEvent(event)

    @property
    def tsa_data(self) -> bytearray:
        return self.model.tile_square_assembly

    @tsa_data.setter
    def tsa_data(self, value: bytearray):
        self.model.tile_square_assembly = value
        self.view.tsa_data = self.tsa_data
        self.tile_square_assembly_changed.emit(self.tsa_data)
        self.view.update()

    def silent_update_tsa_data(self, value: bytearray):
        self.model.tile_square_assembly = value
        self.view.tsa_data = self.tsa_data
        self.view.update()

    @property
    def block_index(self) -> int:
        return self.model.block_index

    @block_index.setter
    def block_index(self, value: int):
        self.model.block_index = value
        self.view.block_index = value
        self.graphics_set_changed.emit(self.block_index)
        self.view.update()
        self.palette_index = value // 0x40

    @property
    def graphics_set(self) -> GraphicsSet:
        return self.model.graphics_set

    @graphics_set.setter
    def graphics_set(self, value: GraphicsSet):
        self.model.graphics_set = value
        self.view.graphics_set = value
        self.graphics_set_changed.emit(self.graphics_set)
        self.view.update()
        if self.pattern_viewer is not None:
            self.pattern_viewer.graphics_set = value

    @property
    def palette_group(self) -> PaletteGroup:
        return self.model.palette_group

    @palette_group.setter
    def palette_group(self, value: PaletteGroup):
        self.model.palette_group = value
        self.view.palette_group = value
        self.palette_group_changed.emit(self.palette_group)
        self.view.update()
        if self.pattern_viewer is not None:
            self.pattern_viewer.palette_group = value

    @property
    def palette_index(self) -> int:
        return self.model.palette_index

    @palette_index.setter
    def palette_index(self, value: int):
        self.model.palette_index = value
        self.view.palette_index = value
        self.palette_index_changed.emit(self.palette_index)
        self.view.update()
        if self.pattern_viewer is not None:
            self.pattern_viewer.palette_index = value

    def _on_pattern_selected(self, x: int, y: int):
        self._last_x = x
        self._last_y = y

        def remove_viewer(*_):
            self.pattern_viewer = None

        if self.pattern_viewer is None:
            self.pattern_viewer = PatternViewer(self, self.graphics_set, self.palette_group, self.block_index // 0x40)
            self.pattern_viewer.pattern_selected.connect(
                lambda ptn: self._on_block_pattern_change(self._last_x, self._last_y, ptn)
            )
            self.pattern_viewer.destroyed.connect(remove_viewer)
            self.pattern_viewer.show()
        else:
            self.pattern_viewer.palette_index = self.block_index // 0x40

    def _on_block_pattern_change(self, x: int, y: int, pattern: int):
        index = self.block_index + 0x100 * (x * 2 + y)
        tsa_data = self.tsa_data
        tsa_data[index] = pattern
        self.tsa_data = tsa_data


class BlockEditorView(QWidget):
    pattern_selected: SignalInstance = Signal(object)  # type: ignore

    def __init__(
        self,
        parent: Optional[QWidget],
        tsa_data: bytearray,
        block_index: int,
        graphics_set: GraphicsSet,
        palette_group: PaletteGroup,
        palette_index: int,
        zoom: int = 16,
    ):
        super().__init__(parent)

        self.tsa_data = tsa_data
        self.block_index = block_index
        self.graphics_set = graphics_set
        self.palette_group = palette_group
        self.palette_index = palette_index
        self.zoom = zoom

    def resizeEvent(self, event: QResizeEvent):
        self.update()

    @property
    def zoom(self) -> int:
        return self._zoom

    @zoom.setter
    def zoom(self, value: int):
        self._zoom = value
        self.setFixedSize(self.block_scale, self.block_scale)

    @property
    def block_scale(self) -> int:
        return Block.WIDTH * self.zoom

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        pos = Position.from_qpoint(event.pos())
        pos.x, pos.y = pos.x // (self.block_scale // 2), pos.y // (self.block_scale // 2)
        self.pattern_selected.emit((pos.x, pos.y))

    def paintEvent(self, event: QPaintEvent):
        painter = QPainter(self)

        bg_color = self.palette_group[0][0]
        painter.setBrush(QBrush(bg_color))
        painter.drawRect(QRect(QPoint(0, 0), self.size()))
        block = Block(self.block_index, self.palette_group, self.graphics_set, self.tsa_data)
        block.draw(painter, 0, 0, self.block_scale)



from collections.abc import Callable
from typing import Optional, Protocol, TypeVar

from PySide6.QtCore import QPoint
from PySide6.QtGui import QAction, QPixmap, Qt
from PySide6.QtWidgets import QScrollArea, QSplitter

from foundry.core.Data import DataProtocol
from foundry.game.gfx.objects.LevelObject import LevelObject
from foundry.game.level import LevelByteData
from foundry.game.level.LevelController import LevelController
from foundry.game.level.LevelRef import LevelRef
from foundry.gui.ContextMenu import ContextMenu
from foundry.gui.JumpCreator import JumpCreator
from foundry.gui.JumpList import JumpList
from foundry.gui.LevelSelector import select_by_world_and_level
from foundry.gui.LevelSizeBar import LevelSizeBar
from foundry.gui.LevelView import LevelView
from foundry.gui.ObjectDropdown import ObjectDropdown
from foundry.gui.ObjectIcon import ObjectButton
from foundry.gui.ObjectIcon import ObjectViewer as ObjectToolbarViewer
from foundry.gui.ObjectList import ObjectList
from foundry.gui.ObjectStatusBar import ObjectStatusBar
from foundry.gui.ObjectToolBar import ObjectToolBar
from foundry.gui.ObjectViewer import ObjectViewer
from foundry.gui.PaletteGroupController import PaletteGroupController
from foundry.gui.PlayerViewer import PlayerViewerController as PlayerViewer
from foundry.gui.SpinnerPanel import SpinnerPanel
from foundry.gui.Toolbar import create_toolbar
from foundry.gui.WarningList import WarningList


class Manager(Protocol):
    @property
    def enabled(self) -> bool:
        ...

    def display_block_viewer(self) -> None:
        ...

    def display_object_viewer(self) -> None:
        ...

    def display_player_viewer(self) -> None:
        ...


T = TypeVar("T")


def require_enabled(function: Callable[..., T]) -> Callable[..., T]:
    def require_enabled_wrapped(self, *args, **kwargs):
        if self.controller is None:
            raise NotImplementedError(f"{self.__class__.__name__} is not enabled")
        return function(self, *args, **kwargs, controller=self.controller)

    return require_enabled_wrapped


class LevelManager:
    def __init__(self, parent):
        self.parent = parent
        self.controller: Optional[LevelController] = None
        self._enabled = False
        self._has_warnings = False

    @property
    def enabled(self) -> bool:
        return self._enabled

    @property
    def has_warnings(self) -> bool:
        return self._has_warnings

    @has_warnings.setter
    def has_warnings(self, value: bool) -> None:
        self._has_warnings
        self.parent.warning_action.setEnabled(value)

    def display_player_viewer(self):
        player_viewer = PlayerViewer(self.parent)
        player_viewer.show()

    def display_object_viewer(self):
        object_viewer = ObjectViewer(parent=self.parent)

        if self.enabled and self.controller is not None:
            object_viewer.set_object_and_graphic_set(
                self.controller.level_ref.level.object_set.number, self.controller.level_ref.level.graphic_set
            )

            if len(self.parent.level_view.get_selected_objects()) == 1:
                selected_object = self.parent.level_view.get_selected_objects()[0]

                if isinstance(selected_object, LevelObject):
                    object_viewer.set_object(selected_object.domain, selected_object.obj_index, selected_object.length)

        object_viewer.show()

    def on_disable(self):
        self._enabled = False

    def on_enable(self):
        self._enabled = True

        level_ref = LevelRef()

        self.controller = LevelController(self.parent, level_ref)

        self.parent.context_menu = ContextMenu(level_ref)
        self.parent.context_menu.triggered.connect(self.parent.on_menu)  # type: ignore
        self.parent.level_view = LevelView(self.parent, level_ref, self.parent.context_menu)
        self.parent.scroll_panel = QScrollArea()
        self.parent.scroll_panel.setWidgetResizable(True)
        self.parent.scroll_panel.setWidget(self.parent.level_view)
        self.parent.setCentralWidget(self.parent.scroll_panel)

        self.parent.spinner_panel = SpinnerPanel(self.parent, level_ref)
        self.parent.spinner_panel.zoom_in_triggered.connect(self.parent.level_view.zoom_in)
        self.parent.spinner_panel.zoom_out_triggered.connect(self.parent.level_view.zoom_out)
        self.parent.spinner_panel.object_change.connect(self.controller.on_spin)

        self.parent.object_list = ObjectList(self.parent, level_ref, self.parent.context_menu)

        self.parent.object_dropdown = ObjectDropdown(self.parent)
        self.parent.object_dropdown.object_selected.connect(self.parent.object_dropdown.select_object)

        self.parent.level_size_bar = LevelSizeBar(self.parent, "Generators", 0, 0)

        def update_level_size_bar(*args, **kwargs):
            self.parent.level_size_bar.current_value = level_ref.level.current_object_size()
            self.parent.level_size_bar.maximum_value = level_ref.level.object_size_on_disk

        level_ref.data_changed.connect(update_level_size_bar)

        self.parent.enemy_size_bar = LevelSizeBar(self.parent, "Enemies/Items", 0, 0)

        def update_enemy_size_bar(*args, **kwargs):
            self.parent.enemy_size_bar.current_value = level_ref.level.current_enemies_size()
            self.parent.enemy_size_bar.maximum_value = level_ref.level.enemy_size_on_disk

        level_ref.data_changed.connect(update_enemy_size_bar)

        self.parent.side_palette = PaletteGroupController(self.parent)
        self.parent.side_palette.palette_group_changed.connect(
            lambda *_: self.controller.update_gui_for_level() if self.controller is not None else None
        )

        self.parent.jump_list = JumpList(self.parent, level_ref)
        self.parent.jump_list.add_jump.connect(self.controller.on_jump_added)
        self.parent.jump_list.edit_jump.connect(self.controller.display_jump_editor)
        self.parent.jump_list.remove_jump.connect(self.controller.on_jump_removed)

        # Toolbar Creation
        splitter = QSplitter(self.parent)
        splitter.setOrientation(Qt.Vertical)

        splitter.addWidget(self.parent.object_list)
        splitter.setStretchFactor(1, 1)
        splitter.addWidget(self.parent.jump_list)
        splitter.addWidget(JumpCreator(parent=self.parent, level_view=self.parent.level_view))

        splitter.setChildrenCollapsible(False)

        create_toolbar(self.parent, "Generator Editor", [self.parent.spinner_panel], Qt.RightToolBarArea)
        create_toolbar(self.parent, "Generator Dropdown", [self.parent.object_dropdown], Qt.RightToolBarArea)
        create_toolbar(self.parent, "Palette", [self.parent.side_palette], Qt.RightToolBarArea)
        create_toolbar(self.parent, "Level Size", [self.parent.level_size_bar], Qt.RightToolBarArea)
        create_toolbar(self.parent, "Enemy Size", [self.parent.enemy_size_bar], Qt.RightToolBarArea)
        create_toolbar(self.parent, "Splitter", [splitter], Qt.RightToolBarArea)

        self.parent.object_toolbar = ObjectToolBar(self.parent)
        self.parent.object_toolbar_viewer = ObjectToolbarViewer(self.parent)

        def set_object_viewer(item):
            self.parent.object_toolbar_viewer.icon = ObjectButton(self.parent.object_toolbar_viewer, item)

        def set_object_viewer_from_selected(items):
            if len(items) == 1:
                self.parent.object_toolbar_viewer.icon = ObjectButton(self.parent.object_toolbar_viewer, items[0])
                self.parent.object_dropdown.select_object(items[0])

        self.parent.level_view.object_created.connect(set_object_viewer)
        self.parent.level_view.object_created.connect(self.parent.object_dropdown.select_object)
        self.parent.object_dropdown.object_selected.connect(set_object_viewer)
        self.parent.level_view.objects_selected.connect(set_object_viewer_from_selected)
        self.parent.object_toolbar.selected.connect(set_object_viewer)
        self.parent.object_toolbar.selected.connect(self.parent.object_dropdown.select_object)

        create_toolbar(self.parent, "Object Viewer", [self.parent.object_toolbar_viewer], Qt.LeftToolBarArea)
        create_toolbar(self.parent, "Object Toolbar", [self.parent.object_toolbar], Qt.LeftToolBarArea)

        # Warning List Creation

        self.parent.warning_list = WarningList(self.parent, level_ref, self.parent.level_view, self.parent.object_list)
        self.parent.warning_list.warnings_updated.connect(lambda value: setattr(self, "has_warnings", value))

        # Status Bar Creation

        self.parent.setStatusBar(ObjectStatusBar(self.parent, level_ref))

    @property
    @require_enabled
    def is_attached(self, controller: LevelController) -> bool:
        return controller.is_attached

    @is_attached.setter
    @require_enabled
    def is_attached(self, value: bool, *, controller: LevelController):
        controller.is_attached = value

    @property
    @require_enabled
    def last_position(self, controller: LevelController) -> tuple[int, int]:
        return controller.last_position

    @property
    @require_enabled
    def actions(self, controller: LevelController) -> list[QAction]:
        return controller.actions

    @require_enabled
    def attach(self, generator_pointer: int, enemy_pointer: int, *, controller: LevelController):
        controller.attach(generator_pointer, enemy_pointer)

    @require_enabled
    def to_data(self, controller: LevelController) -> list[DataProtocol]:
        return controller.to_data()

    @require_enabled
    def load_m3l(self, m3l_data: bytearray, pathname: str, *, controller: LevelController):
        controller.load_m3l(m3l_data, pathname)

    @require_enabled
    def to_m3l(self, controller: LevelController) -> bytearray:
        return controller.to_m3l()

    @require_enabled
    def refresh(self, controller: LevelController):
        controller.refresh()

    @require_enabled
    def update(self, controller: LevelController):
        controller.update()

    @require_enabled
    def force_select(self, world: int, level: int, *, controller: LevelController):
        selection = select_by_world_and_level(world, level)
        controller.update_level(
            selection.display_information.name if selection.display_information.name is not None else "",
            selection.generator_pointer - 9,
            selection.enemy_pointer,
            selection.tileset,
        )

    @require_enabled
    def on_select(self, controller: LevelController):
        return controller.on_select()

    @property
    @require_enabled
    def safe_to_change(self, controller: LevelController) -> bool:
        return controller.safe_to_change

    @safe_to_change.setter
    @require_enabled
    def safe_to_change(self, value: bool, *, controller: LevelController):
        controller.safe_to_change = value

    @property
    @require_enabled
    def stable_changes(self, controller: LevelController) -> tuple[bool, str, str]:
        return controller.stable_changes

    @property
    def state(self) -> LevelByteData:
        if self.controller is None:
            raise NotImplementedError(f"{self.__class__.__name__} is not enabled")
        return self.controller.level_ref.state

    def do(self, new_state: LevelByteData) -> LevelByteData:
        if self.controller is None:
            raise NotImplementedError(f"{self.__class__.__name__} is not enabled")
        return self.controller.level_ref.do(new_state)

    @property
    def can_undo(self) -> bool:
        return self.controller is not None and self.controller.level_ref.can_undo

    def undo(self) -> LevelByteData:
        if self.controller is None:
            raise NotImplementedError(f"{self.__class__.__name__} is not enabled")
        return self.controller.level_ref.undo()

    @property
    def can_redo(self) -> bool:
        return self.controller is not None and self.controller.level_ref.can_redo

    def redo(self) -> LevelByteData:
        if self.controller is None:
            raise NotImplementedError(f"{self.__class__.__name__} is not enabled")
        return self.controller.level_ref.redo()

    @require_enabled
    def delete(self, controller: LevelController):
        controller.delete()

    @require_enabled
    def create_object_from_suggestion(self, position: tuple[int, int], *, controller: LevelController):
        if controller.suggested_object != -1:
            controller.place_object_from_dropdown(position)
        else:
            controller.create_object_at(*position)

    @require_enabled
    def to_foreground(self, controller: LevelController):
        controller.to_foreground()

    @require_enabled
    def to_background(self, controller: LevelController):
        controller.to_background()

    @property
    @require_enabled
    def title_suggestion(self, controller: LevelController) -> str:
        return controller.title_suggestion

    @property
    @require_enabled
    def screenshot(self, controller: LevelController) -> QPixmap:
        return controller.screenshot

    @require_enabled
    def cut(self, controller: LevelController):
        controller.cut()

    @require_enabled
    def copy(self, controller: LevelController):
        controller.copy()

    @require_enabled
    def paste(self, x: Optional[int] = None, y: Optional[int] = None, *, controller: LevelController):
        return controller.paste(x, y)

    @require_enabled
    def zoom_in(self, controller: LevelController):
        controller.zoom_in()

    @require_enabled
    def zoom_out(self, controller: LevelController):
        controller.zoom_out()

    @require_enabled
    def warp_to_alternative(self, controller: LevelController):
        controller.warp_to_alternative()

    @require_enabled
    def display_warnings(self, controller: LevelController):
        controller.display_warnings()

    @require_enabled
    def select_all(self, controller: LevelController):
        controller.select_all()

    @require_enabled
    def focus_selected(self, controller: LevelController):
        controller.focus_selected()

    @require_enabled
    def middle_mouse_release(self, position: QPoint, *, controller: LevelController):
        controller.middle_mouse_release(position)

    @require_enabled
    def display_block_viewer(self, controller: LevelController):
        controller.display_block_viewer()

    @require_enabled
    def display_palette_viewer(self, controller: LevelController):
        controller.display_palette_viewer()

    @require_enabled
    def display_autoscroll_editor(self, controller: LevelController):
        controller.display_autoscroll_editor()

    @require_enabled
    def display_header_editor(self, controller: LevelController):
        controller.display_header_editor()

    @require_enabled
    def display_jump_editor(self, controller: LevelController):
        controller.display_jump_editor()



from PySide6.QtGui import QKeyEvent, Qt
from PySide6.QtWidgets import QDialog


class CustomDialog(QDialog):
    def __init__(self, parent, title="Title"):
        super(CustomDialog, self).__init__(parent)

        self.setWindowTitle(title)
        self.setWindowFlag(Qt.WindowStaysOnTopHint)

    def keyPressEvent(self, event: QKeyEvent):
        if event.key() == Qt.Key_Escape:
            self.on_exit()

    def on_exit(self):
        self.hide()

    def closeEvent(self, event):
        self.on_exit()



from foundry.smb3parse.constants import TILESET_LEVEL_OFFSET
from foundry.smb3parse.levels import (
    DEFAULT_HORIZONTAL_HEIGHT,
    DEFAULT_VERTICAL_WIDTH,
    ENEMY_BASE_OFFSET,
    HEADER_LENGTH,
    LEVEL_BASE_OFFSET,
    LEVEL_LENGTH_INTERVAL,
    LEVEL_MIN_LENGTH,
)
from foundry.smb3parse.objects.object_set import assert_valid_object_set_number

MARIO_X_POSITIONS = [0x18, 0x70, 0xD8, 0x80]  # 0x10249
MARIO_Y_POSITIONS = [0x17, 0x04, 0x00, 0x14, 0x07, 0x0B, 0x0F, 0x18]  # 0x3D7A0 + 0x3D7A8


class LevelHeader:
    def __init__(self, header_bytes: bytearray, object_set_number: int):
        if len(header_bytes) != HEADER_LENGTH:
            raise ValueError(f"A level header is made up of {HEADER_LENGTH} bytes, but {len(header_bytes)} were given.")

        assert_valid_object_set_number(object_set_number)

        self._object_set_number = object_set_number

        self.data = header_bytes

        self.start_y_index = (self.data[4] & 0b1110_0000) >> 5

        self.screens = self.data[4] & 0b0000_1111
        self.length = LEVEL_MIN_LENGTH + self.screens * LEVEL_LENGTH_INTERVAL
        self.width = self.length
        self.height = DEFAULT_HORIZONTAL_HEIGHT

        self.start_x_index = (self.data[5] & 0b0110_0000) >> 5

        self.enemy_palette_index = (self.data[5] & 0b0001_1000) >> 3
        self.object_palette_index = self.data[5] & 0b0000_0111

        self.pipe_ends_level = not (self.data[6] & 0b1000_0000)
        self.scroll_type_index = (self.data[6] & 0b0110_0000) >> 5
        self.is_vertical = self.data[6] & 0b0001_0000

        if self.is_vertical:
            self.height = self.length
            self.width = DEFAULT_VERTICAL_WIDTH

        self.jump_object_set_number = self.data[6] & 0b0000_1111  # for indexing purposes

        self.start_action = (self.data[7] & 0b1110_0000) >> 5

        self.graphic_set_index = self.data[7] & 0b0001_1111

        self.time_index = (self.data[8] & 0b1100_0000) >> 6

        self.music_index = self.data[8] & 0b0000_1111

        self.jump_level_address = (
            (self.data[1] << 8) + self.data[0] + LEVEL_BASE_OFFSET + TILESET_LEVEL_OFFSET[self.jump_object_set_number]
        )
        self.jump_enemy_address = (self.data[3] << 8) + self.data[2] + ENEMY_BASE_OFFSET

    def mario_position(self):
        x = MARIO_X_POSITIONS[self.start_x_index] >> 4
        y = MARIO_Y_POSITIONS[self.start_y_index]

        if self.is_vertical:
            y += (self.screens - 1) * 15  # TODO: Why?

        return x, y



from typing import Optional

from PySide6.QtGui import Qt
from PySide6.QtWidgets import QMainWindow, QToolBar, QWidget


def create_toolbar(
    parent: QMainWindow, name: str, widgets: Optional[list[QWidget]] = None, area: Optional[Qt.ToolBarArea] = None
) -> QToolBar:
    """
    Creates a QToolbar with the default arguments for this repository.

    Theses arguments include context menu policy, size policy, orientation, and floatable.

    Parameters
    ----------
    parent : QMainWindow
        The parent of the toolbar
    name : str
        The name of the toolbar
    widgets: Optional[list[QWidget]]
        The widgets to attach to the toolbar

    Returns
    -------
    QToolBar
        with the default arguments
    """

    toolbar = QToolBar(name, parent)
    toolbar.setContextMenuPolicy(Qt.PreventContextMenu)
    toolbar.setOrientation(Qt.Horizontal)
    toolbar.setFloatable(True)
    toolbar.setAllowedAreas(Qt.LeftToolBarArea | Qt.RightToolBarArea)

    if widgets is not None:
        for widget in widgets:
            toolbar.addWidget(widget)

    if area is not None:
        parent.addToolBar(area, toolbar)

    return toolbar



from typing import Protocol

from attr import attrs

from foundry.core.Position import Position, PositionProtocol


class EnemyProtocol(Protocol):
    """
    A representation of an enemy inside a level.

    Attributes
    ----------
    type: int
        The type of enemy
    position: PositionProtocol
        The position of the enemy
    """

    type: int
    position: PositionProtocol

    def __bytes__(self) -> bytes:
        ...

    @classmethod
    def from_bytes(cls, data: bytes):
        ...


@attrs(slots=True, auto_attribs=True)
class Enemy:
    """
    A representation of an enemy inside a level, that uses ``attrs`` to create a basic implementation.

    Attributes
    ----------
    type: int
        The type of enemy
    position: PositionProtocol
        The position of the enemy
    """

    type: int
    position: PositionProtocol

    def __bytes__(self) -> bytes:
        return bytes([self.type, self.position.x, self.position.y])

    @classmethod
    def from_bytes(cls, data: bytes):
        return cls(data[0], Position(data[1], data[2]))



from PySide6.QtCore import QRect, QSize
from PySide6.QtGui import QColor, QImage, QPainter, Qt

from foundry.core.Position import Position, PositionProtocol
from foundry.game.EnemyDefinitions import (
    EnemyDefinition,
    GeneratorType,
    get_enemy_metadata,
)
from foundry.game.gfx.drawable import apply_selection_overlay
from foundry.game.gfx.drawable.Block import Block
from foundry.game.gfx.drawable.Sprite import get_sprite
from foundry.game.gfx.GraphicsSet import GraphicalPage, GraphicsSet
from foundry.game.gfx.objects.Enemy import Enemy
from foundry.game.gfx.objects.ObjectLike import ObjectLike
from foundry.game.gfx.Palette import PaletteGroup
from foundry.smb3parse.objects.object_set import ENEMY_ITEM_GRAPHICS_SET

MASK_COLOR = [0xFF, 0x33, 0xFF]


class EnemyObject(ObjectLike):
    def __init__(self, data, png_data, palette_group: PaletteGroup):
        super().__init__()
        self.enemy = Enemy.from_bytes(data)

        self.graphics_set = GraphicsSet.from_tileset(ENEMY_ITEM_GRAPHICS_SET)
        self.palette_group = palette_group

        self.png_data = png_data

        self.selected = False

        self._setup()

    @property
    def definition(self) -> EnemyDefinition:
        return get_enemy_metadata().__root__[self.obj_index]

    @property
    def rect(self):
        bmp_width = (
            self.definition.bmp_width
            if not GeneratorType.SINGLE_SPRITE_OBJECT == self.definition.orientation
            else self.definition.bmp_width // 2
        )
        width = self.definition.rect_width if self.definition.rect_width != 0 else bmp_width
        height = self.definition.rect_height if self.definition.rect_height != 0 else self.definition.bmp_height

        return QRect(
            self.position.x - self.definition.rect_x_offset,
            self.position.y - self.definition.rect_y_offset,
            width,
            height,
        )

    def _setup(self):
        obj_def = get_enemy_metadata().__root__[self.obj_index]

        if GeneratorType.SINGLE_SPRITE_OBJECT == obj_def.orientation:
            self.graphics_set = GraphicsSet(tuple(GraphicalPage(page) for page in obj_def.pages))

        self.name = obj_def.description

        self.width = obj_def.bmp_width
        self.height = obj_def.bmp_height

        self._render(obj_def)

    def _render(self, obj_def: EnemyDefinition):
        if not GeneratorType.SINGLE_SPRITE_OBJECT == obj_def.orientation:
            self._render_blocks(obj_def)
        else:
            self._render_sprites(obj_def)

    def _render_sprites(self, obj_def: EnemyDefinition):
        self.sprites = obj_def.sprites

    def _render_blocks(self, obj_def: EnemyDefinition):
        self.blocks = []

        block_ids = obj_def.blocks

        for block_id in block_ids:
            x = (block_id % 64) * Block.WIDTH
            y = (block_id // 64) * Block.WIDTH

            self.blocks.append(self.png_data.copy(QRect(x, y, Block.WIDTH, Block.HEIGHT)))

    def render(self):
        # nothing to re-render since enemies are just copied over
        pass

    def draw(self, painter: QPainter, block_length, transparency, *, is_icon=False):
        if not GeneratorType.SINGLE_SPRITE_OBJECT == self.definition.orientation:
            self.draw_blocks(painter, block_length, is_icon)
        else:
            self.draw_sprites(painter, block_length // 2, transparency, is_icon)

    def draw_sprites(self, painter: QPainter, scale_factor, transparency, is_icon):
        for i, sprite_info in enumerate(self.sprites):
            if sprite_info.index < 0:
                continue

            x = (self.position.x * 2) + (i % self.width) if not is_icon else (i % self.width)
            y = self.position.y + (i // self.width) if not is_icon else (i // self.width)
            x += sprite_info.x_offset / 16
            y -= sprite_info.y_offset / 16

            if is_icon:
                definition = get_enemy_metadata().__root__[self.obj_index]
                x_offset, y_offset = definition.suggested_icon_x_offset, definition.suggested_icon_y_offset
                x += x_offset / 16
                y -= y_offset / 16
            if not is_icon:
                y_offset = self.height - 1
                y -= y_offset

            sprite = get_sprite(
                sprite_info.index,
                tuple(tuple(c for c in pal) for pal in self.palette_group),
                sprite_info.palette_index,
                self.graphics_set,
                sprite_info.horizontal_mirror,
                sprite_info.vertical_mirror,
            )
            sprite.draw(
                painter,
                x * scale_factor,
                y * scale_factor * 2,
                scale_factor,
                scale_factor * 2,
                self.selected,
                transparency,
            )

    def draw_blocks(self, painter: QPainter, block_length, is_icon):
        for i, image in enumerate(self.blocks):
            x = self.position.x + (i % self.width) if not is_icon else (i % self.width)
            y = self.position.y + (i // self.width) if not is_icon else (i // self.width)

            if is_icon:
                definition = get_enemy_metadata().__root__[self.obj_index]
                x_offset, y_offset = definition.suggested_icon_x_offset, definition.suggested_icon_y_offset
                x -= x_offset
            if not is_icon:
                y_offset = self.height - 1
                y -= y_offset

            block = image.copy()

            mask = block.createMaskFromColor(QColor(*MASK_COLOR).rgb(), Qt.MaskOutColor)
            block.setAlphaChannel(mask)

            # todo better effect
            if self.selected:
                apply_selection_overlay(block, mask)

            if block_length != Block.SIDE_LENGTH:
                block = block.scaled(block_length, block_length)

            painter.drawImage(x * block_length, y * block_length, block)

    def get_status_info(self):
        return [("Name", self.name), ("X", self.position.x), ("Y", self.position.y)]

    def __contains__(self, item):
        x, y = item

        return self.point_in(x, y)

    def point_in(self, x, y):
        return self.rect.contains(x, y)

    def move_by(self, dx, dy):
        self.position = Position(self.position.x + dx, self.position.y + dy)

    @property
    def position(self) -> PositionProtocol:
        return self.enemy.position

    @position.setter
    def position(self, position: PositionProtocol):
        self.enemy.position = Position(max(0, position.x), max(0, position.y))

    def resize_by(self, dx, dy):
        pass

    @property
    def obj_index(self):
        return self.enemy.type

    @property
    def type(self):
        return self.enemy.type

    def change_type(self, new_type):
        self.enemy.type = new_type

        self._setup()

    def increment_type(self):
        self.enemy.type = min(0xFF, self.obj_index + 1)

        self._setup()

    def decrement_type(self):
        self.enemy.type = max(0, self.obj_index - 1)

        self._setup()

    def to_bytes(self):
        return bytes(self.enemy)

    def as_image(self) -> QImage:
        definition = get_enemy_metadata().__root__[self.obj_index]
        width, height = definition.suggested_icon_width * 16, definition.suggested_icon_height * 16

        image = QImage(QSize(width, height), QImage.Format_RGBA8888)
        image.fill(QColor(0, 0, 0, 0))

        painter = QPainter(image)

        self.draw(painter, Block.SIDE_LENGTH, True, is_icon=True)

        return image

    def __str__(self):
        return f"{self.name} at {self.position.x}, {self.position.y}"

    def __repr__(self):
        return f"EnemyObject: {self}"



from PySide6.QtWidgets import QMessageBox


class AutoSaveDialog(QMessageBox):
    def __init__(self):
        super(AutoSaveDialog, self).__init__()

        self.setWindowTitle("Rom was recovered")
        self.setText("We found an auto saved ROM from the last session. Do you want to open it?")
        self.setIcon(QMessageBox.Warning)

        self.use_auto_save_button = self.addButton("Load Auto Save", QMessageBox.AcceptRole)
        self.setDefaultButton(self.use_auto_save_button)

        self.discard_rom_button = self.addButton("Discard Auto Save", QMessageBox.DestructiveRole)



from json import loads
from typing import Optional

from PySide6.QtWidgets import QWidget

from foundry import jump_creator_flags_path
from foundry.gui.HeaderEditor import HeaderEditor
from foundry.gui.LevelView import LevelView, undoable
from foundry.gui.util import setup_layout


class JumpCreator(QWidget):
    """
    A menu item used to create and modify jumps inside the level editor.
    """

    def __init__(self, level_view: LevelView, parent: Optional[QWidget] = None):
        super().__init__(parent)
        self.level_view = level_view
        self.level_ref = level_view.level_ref

        with open(jump_creator_flags_path, "r") as data:
            flags = loads(data.read())
        setup_layout(self, flags)

    @undoable
    def add_jump(self):
        self.level_view.add_jump()

    def show_jump_dest(self):
        header_editor = HeaderEditor(self, self.level_ref)
        header_editor.tab_widget.setCurrentIndex(3)
        header_editor.exec_()



from typing import Optional

from PySide6.QtCore import Signal
from PySide6.QtGui import Qt
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from foundry.game.gfx.GraphicsSet import GRAPHIC_SET_NAMES
from foundry.game.level.Level import Level
from foundry.game.level.LevelRef import LevelRef
from foundry.gui.CustomDialog import CustomDialog
from foundry.gui.LevelSelector import OBJECT_SET_ITEMS, LevelSelector
from foundry.gui.Spinner import Spinner
from foundry.smb3parse.levels.level_header import MARIO_X_POSITIONS, MARIO_Y_POSITIONS

LEVEL_LENGTHS = [0x10 * (i + 1) for i in range(0, 2 ** 4)]
STR_LEVEL_LENGTHS = [f"{length - 1:0=#4X} / {length} Blocks".replace("X", "x") for length in LEVEL_LENGTHS]

STR_X_POSITIONS = [f"{position >> 4}. Block ({position:0=#4X})".replace("X", "x") for position in MARIO_X_POSITIONS]

STR_Y_POSITIONS = [f"{position}. Block ({position:0=#4X})".replace("X", "x") for position in MARIO_Y_POSITIONS]

ACTIONS = [
    "None",
    "Sliding",
    "Out of pipe ↑",
    "Out of pipe ↓",
    "Out of pipe ←",
    "Out of pipe →",
    "Running and climbing up ship",
    "Ship auto scrolling",
]

MUSIC_ITEMS = [
    "Plain level",
    "Underground",
    "Water level",
    "Fortress",
    "Boss",
    "Ship",
    "Battle",
    "P-Switch/Mushroom house (1)",
    "Hilly level",
    "Castle room",
    "Clouds/Sky",
    "P-Switch/Mushroom house (2)",
    "No music",
    "P-Switch/Mushroom house (1)",
    "No music",
    "World 7 map",
]

TIMES = ["300", "400", "200", "Unlimited"]

SCROLL_DIRECTIONS = [
    "Locked, unless climbing/flying",
    "Free vertical scrolling",
    "Locked 'by start coordinates'?",
    "Shouldn't appear in game, do not use.",
]


SPINNER_MAX_VALUE = 0x0F_FF_FF


class HeaderEditor(CustomDialog):
    header_change: Signal = Signal()

    def __init__(self, parent: Optional[QWidget], level_ref: LevelRef):
        super(HeaderEditor, self).__init__(parent, "Level Header Editor")

        self.level: Level = level_ref.level
        assert self.level is not None

        # enables undo
        self.header_change.connect(level_ref.save_level_state)

        main_layout = QVBoxLayout(self)

        self.tab_widget = QTabWidget(self)
        main_layout.addWidget(self.tab_widget)

        # level settings

        self.length_dropdown = QComboBox()
        if self.level.is_vertical:
            self.length_dropdown.addItems(STR_LEVEL_LENGTHS)
        else:
            self.length_dropdown.addItems(STR_LEVEL_LENGTHS[:-1])

        self.length_dropdown.activated.connect(self.on_combo)

        self.music_dropdown = QComboBox()
        self.music_dropdown.addItems(MUSIC_ITEMS)
        self.music_dropdown.activated.connect(self.on_combo)

        self.time_dropdown = QComboBox()
        self.time_dropdown.addItems(TIMES)
        self.time_dropdown.activated.connect(self.on_combo)

        self.v_scroll_direction_dropdown = QComboBox()
        self.v_scroll_direction_dropdown.addItems(SCROLL_DIRECTIONS)
        self.v_scroll_direction_dropdown.activated.connect(self.on_combo)

        self.level_is_vertical_cb = QCheckBox("Level is Vertical")
        self.level_is_vertical_cb.clicked.connect(self.on_check_box)

        self.pipe_ends_level_cb = QCheckBox("Pipe ends Level")
        self.pipe_ends_level_cb.clicked.connect(self.on_check_box)

        check_box_layout = QHBoxLayout()
        check_box_layout.setContentsMargins(0, 0, 0, 0)
        check_box_layout.addWidget(self.level_is_vertical_cb)
        check_box_layout.addWidget(self.pipe_ends_level_cb)

        check_box_widget = QWidget()
        check_box_widget.setLayout(check_box_layout)

        form = QFormLayout()
        form.setFormAlignment(Qt.AlignCenter)

        form.addRow("Level length: ", self.length_dropdown)
        form.addRow("Music: ", self.music_dropdown)
        form.addRow("Time: ", self.time_dropdown)
        form.addRow("Scroll direction: ", self.v_scroll_direction_dropdown)

        form.addWidget(check_box_widget)

        widget = QWidget()
        widget.setLayout(form)

        self.tab_widget.addTab(widget, "Level")

        # player settings

        self.x_position_dropdown = QComboBox()
        self.x_position_dropdown.addItems(STR_X_POSITIONS)
        self.x_position_dropdown.activated.connect(self.on_combo)

        self.y_position_dropdown = QComboBox()
        self.y_position_dropdown.addItems(STR_Y_POSITIONS)
        self.y_position_dropdown.activated.connect(self.on_combo)

        self.action_dropdown = QComboBox()
        self.action_dropdown.addItems(ACTIONS)
        self.action_dropdown.activated.connect(self.on_combo)

        form = QFormLayout()
        form.setFormAlignment(Qt.AlignCenter)

        form.addRow("Starting X: ", self.x_position_dropdown)
        form.addRow("Starting Y: ", self.y_position_dropdown)
        form.addRow("Action: ", self.action_dropdown)

        widget = QWidget()
        widget.setLayout(form)

        self.tab_widget.addTab(widget, "Mario")

        # graphic settings

        self.object_palette_spinner = Spinner(self, maximum=7)
        self.object_palette_spinner.valueChanged.connect(self.on_spin)

        self.enemy_palette_spinner = Spinner(self, maximum=3)
        self.enemy_palette_spinner.valueChanged.connect(self.on_spin)

        self.graphic_set_dropdown = QComboBox()
        self.graphic_set_dropdown.addItems(GRAPHIC_SET_NAMES)
        self.graphic_set_dropdown.activated.connect(self.on_combo)

        form = QFormLayout()
        form.setFormAlignment(Qt.AlignCenter)

        form.addRow("Object Palette: ", self.object_palette_spinner)
        form.addRow("Enemy Palette: ", self.enemy_palette_spinner)
        form.addRow("Graphic Set: ", self.graphic_set_dropdown)

        widget = QWidget()
        widget.setLayout(form)

        self.tab_widget.addTab(widget, "Graphics")

        # next area settings

        self.level_pointer_spinner = Spinner(self, maximum=SPINNER_MAX_VALUE)
        self.level_pointer_spinner.valueChanged.connect(self.on_spin)
        self.enemy_pointer_spinner = Spinner(self, maximum=SPINNER_MAX_VALUE)
        self.enemy_pointer_spinner.valueChanged.connect(self.on_spin)

        self.next_area_object_set_dropdown = QComboBox()
        self.next_area_object_set_dropdown.addItems(OBJECT_SET_ITEMS)
        self.next_area_object_set_dropdown.activated.connect(self.on_combo)

        level_select_button = QPushButton("Set from Level")
        level_select_button.pressed.connect(self._set_jump_destination)

        form = QFormLayout()
        form.setFormAlignment(Qt.AlignCenter)

        form.addRow("Address of Objects: ", self.level_pointer_spinner)
        form.addRow("Address of Enemies: ", self.enemy_pointer_spinner)
        form.addRow("Object Set: ", self.next_area_object_set_dropdown)

        form.addRow(QLabel(""))
        form.addRow(level_select_button)

        widget = QWidget()
        widget.setLayout(form)

        self.tab_widget.addTab(widget, "Warping")

        self.header_bytes_label = QLabel()

        main_layout.addWidget(self.header_bytes_label, alignment=Qt.AlignCenter)

        self.update()

    def update(self):
        length_index = LEVEL_LENGTHS.index(self.level.length)

        self.length_dropdown.setCurrentIndex(length_index)
        self.music_dropdown.setCurrentIndex(self.level.music_index)
        self.time_dropdown.setCurrentIndex(self.level.time_index)
        self.v_scroll_direction_dropdown.setCurrentIndex(self.level.scroll_type)
        self.level_is_vertical_cb.setChecked(self.level.is_vertical)
        self.pipe_ends_level_cb.setChecked(self.level.pipe_ends_level)

        self.x_position_dropdown.setCurrentIndex(self.level.start_x_index)
        self.y_position_dropdown.setCurrentIndex(self.level.start_y_index)
        self.action_dropdown.setCurrentIndex(self.level.start_action)

        self.object_palette_spinner.setValue(self.level.object_palette_index)
        self.enemy_palette_spinner.setValue(self.level.enemy_palette_index)
        self.graphic_set_dropdown.setCurrentIndex(self.level.graphic_set)

        self.level_pointer_spinner.setValue(self.level.next_area_objects)
        self.enemy_pointer_spinner.setValue(self.level.next_area_enemies)
        self.next_area_object_set_dropdown.setCurrentIndex(self.level.next_area_object_set)

        self.header_bytes_label.setText(" ".join(f"{number:0=#4X}"[2:] for number in self.level.header_bytes))

    def _set_jump_destination(self):
        level_selector = LevelSelector(self)
        level_was_selected = level_selector.exec_() == QDialog.Accepted

        if not level_was_selected:
            return

        self.next_area_object_set_dropdown.setCurrentIndex(level_selector.object_set)
        self.level.next_area_object_set = level_selector.object_set

        self.level_pointer_spinner.setValue(level_selector.object_data_offset)
        self.level.next_area_objects = level_selector.object_data_offset

        self.enemy_pointer_spinner.setValue(level_selector.enemy_data_offset)
        self.level.next_area_enemies = level_selector.enemy_data_offset - 1

        self.header_change.emit()

    def on_spin(self, _):
        if self.level is None:
            return

        spinner = self.sender()

        self.level.data_changed.connect(self.header_change.emit)

        if spinner == self.object_palette_spinner:
            new_index = self.object_palette_spinner.value()
            self.level.object_palette_index = new_index

        elif spinner == self.enemy_palette_spinner:
            new_index = self.enemy_palette_spinner.value()
            self.level.enemy_palette_index = new_index

        elif spinner == self.level_pointer_spinner:
            new_offset = self.level_pointer_spinner.value()
            self.level.next_area_objects = new_offset

        elif spinner == self.enemy_pointer_spinner:
            new_offset = self.enemy_pointer_spinner.value()
            self.level.next_area_enemies = new_offset

        self.level.data_changed.disconnect()

        self.update()

    def on_combo(self, _):
        dropdown = self.sender()

        self.level.data_changed.connect(self.header_change.emit)

        if dropdown == self.length_dropdown:
            new_length = LEVEL_LENGTHS[self.length_dropdown.currentIndex()]
            self.level.length = new_length

        elif dropdown == self.music_dropdown:
            new_music = self.music_dropdown.currentIndex()
            self.level.music_index = new_music

        elif dropdown == self.time_dropdown:
            new_time = self.time_dropdown.currentIndex()
            self.level.time_index = new_time

        elif dropdown == self.v_scroll_direction_dropdown:
            new_scroll = self.v_scroll_direction_dropdown.currentIndex()
            self.level.scroll_type = new_scroll

        elif dropdown == self.x_position_dropdown:
            new_x = self.x_position_dropdown.currentIndex()
            self.level.start_x_index = new_x

        elif dropdown == self.y_position_dropdown:
            new_y = self.y_position_dropdown.currentIndex()
            self.level.start_y_index = new_y

        elif dropdown == self.action_dropdown:
            new_action = self.action_dropdown.currentIndex()
            self.level.start_action = new_action

        elif dropdown == self.graphic_set_dropdown:
            new_gfx_set = self.graphic_set_dropdown.currentIndex()
            self.level.graphic_set = new_gfx_set

        elif dropdown == self.next_area_object_set_dropdown:
            new_object_set = self.next_area_object_set_dropdown.currentIndex()
            self.level.next_area_object_set = new_object_set

        self.level.data_changed.disconnect()

        self.update()

    def on_check_box(self, _):
        checkbox = self.sender()

        self.level.data_changed.connect(self.header_change.emit)

        if checkbox == self.pipe_ends_level_cb:
            self.level.pipe_ends_level = self.pipe_ends_level_cb.isChecked()
        elif checkbox == self.level_is_vertical_cb:
            self.level.is_vertical = self.level_is_vertical_cb.isChecked()

            self.length_dropdown.clear()
            if self.level.is_vertical:
                self.length_dropdown.addItems(STR_LEVEL_LENGTHS)
            else:
                self.length_dropdown.addItems(STR_LEVEL_LENGTHS[:-1])

        self.level.data_changed.disconnect()

        self.update()



from itertools import chain
from typing import Protocol

from attr import attrs

from foundry.game.File import ROM
from foundry.smb3parse.constants import Level_BG_Pages1, Level_BG_Pages2

CHR_ROM_OFFSET = 0x40010
CHR_ROM_SEGMENT_SIZE = 0x400

WORLD_MAP = 0
SPADE_ROULETTE = 16
N_SPADE = 17
VS_2P = 18

BG_PAGE_COUNT = Level_BG_Pages2 - Level_BG_Pages1  # 23 in stock rom

GRAPHIC_SET_NAMES = [
    "Mario graphics (1)",
    "Plain",
    "Dungeon",
    "Underground (1)",
    "Sky",
    "Pipe/Water (1, Piranha Plant)",
    "Pipe/Water (2, Water)",
    "Mushroom house (1)",
    "Pipe/Water (3, Pipe)",
    "Desert",
    "Ship",
    "Giant",
    "Ice",
    "Clouds",
    "Underground (2)",
    "Spade bonus room",
    "Spade bonus",
    "Mushroom house (2)",
    "Pipe/Water (4)",
    "Hills",
    "Plain 2",
    "Tank",
    "Castle",
    "Mario graphics (2)",
    "Animated graphics (1)",
    "Animated graphics (2)",
    "Animated graphics (3)",
    "Animated graphics (4)",
    "Animated graphics (P-Switch)",
    "Game font/Course Clear graphics",
    "Animated graphics (5)",
    "Animated graphics (6)",
]


class GraphicalPageProtocol(Protocol):
    """
    A representation of a single page of graphics inside the ROM.

    Attributes
    ----------
    index: int
        The index of the graphical page into the ROM.
    """

    index: int

    def __bytes__(self) -> bytes:
        ...


@attrs(slots=True, auto_attribs=True, frozen=True, eq=True, hash=True)
class GraphicalPage:
    """
    A representation of a single page of graphics inside the ROM, that uses ``attrs`` to create a
    basic implementation.

    Attributes
    ----------
    index: int
        The index of the graphical page into the ROM.
    """

    index: int

    @property
    def rom_offset(self) -> int:
        return CHR_ROM_OFFSET + self.index * CHR_ROM_SEGMENT_SIZE

    def __bytes__(self) -> bytes:
        return bytes(ROM().bulk_read(CHR_ROM_SEGMENT_SIZE, self.rom_offset))


class GraphicsSetProtocol(Protocol):
    """
    A representation of a series of graphical pages inside the ROM.

    Attributes
    ----------
    pages: tuple[GraphicalPageProtocol, ...]
        The pages that compose the graphical set.
    """

    pages: tuple[GraphicalPageProtocol, ...]

    def __bytes__(self) -> bytes:
        ...


@attrs(slots=True, auto_attribs=True, frozen=True, eq=True, hash=True)
class GraphicsSet:
    """
    A representation of a series of graphical pages inside the ROM, that uses ``attrs`` to create
    a basic implementation.

    Attributes
    ----------
    pages: tuple[GraphicalPageProtocol, ...]
        The pages that compose the graphical set.
    """

    pages: tuple[GraphicalPageProtocol, ...]

    def __bytes__(self) -> bytes:
        return bytes(chain.from_iterable([bytes(page) for page in self.pages]))

    @classmethod
    def from_tileset(cls, index: int):
        cls.number = index
        if index == WORLD_MAP:
            return cls(
                (
                    GraphicalPage(0x14),
                    GraphicalPage(0x15),
                    GraphicalPage(0x16),
                    GraphicalPage(0x17),
                    GraphicalPage(0x20),
                    GraphicalPage(0x21),
                    GraphicalPage(0x22),
                    GraphicalPage(0x23),
                )
            )
        if index not in range(BG_PAGE_COUNT):
            return cls(
                (GraphicalPage(index), GraphicalPage(index + 1), GraphicalPage(index + 2), GraphicalPage(index + 3))
            )

        graphic_page_index_1 = ROM().bulk_read(BG_PAGE_COUNT, Level_BG_Pages1)
        graphic_page_index_2 = ROM().bulk_read(BG_PAGE_COUNT, Level_BG_Pages2)
        pages = [
            GraphicalPage(graphic_page_index_1[index]),
            GraphicalPage(graphic_page_index_1[index] + 1),
            GraphicalPage(graphic_page_index_2[index]),
            GraphicalPage(graphic_page_index_2[index] + 1),
        ]

        if index == SPADE_ROULETTE:
            pages.extend([GraphicalPage(0x20), GraphicalPage(0x21), GraphicalPage(0x22), GraphicalPage(0x23)])
        elif index == N_SPADE:
            pages.extend([GraphicalPage(0x28), GraphicalPage(0x29), GraphicalPage(0x5A), GraphicalPage(0x31)])
        elif index == VS_2P:
            pages.extend([GraphicalPage(0x04), GraphicalPage(0x05), GraphicalPage(0x06), GraphicalPage(0x07)])
        else:
            pages.extend([GraphicalPage(0x00), GraphicalPage(0x00), GraphicalPage(0x00), GraphicalPage(0x00)])

        return cls(tuple(pages))



"""
Describes all objects, that are part of a level, i. e. platforms, enemies, items, jumps, auto scroll objects etc.
"""
from abc import ABC
from typing import Optional, Protocol

import attr

from foundry.core.Position import Position, PositionProtocol

MIN_DOMAIN = 0
MAX_DOMAIN = 7
MIN_Y_VALUE = 0
MAX_Y_VALUE = 27
MIN_ID_VALUE = 0
MAX_ID_VALUE = 0xFF
MIN_X_VALUE = 0
MAX_X_VALUE = 0xFF
MIN_ADDITIONAL_LENGTH = 0
MAX_ADDITIONAL_LENGTH = 0xFF

MAX_ENEMY_ITEM_ID = 0xEC


class LevelComponentProtocol(Protocol):
    """
    An outline of any object that composes an element inside the game.

    Attributes
    ----------
    domain: int
        The domain or set of component that the component is housed inside.
    index: int
        The specific byte of data that determines a limited amount of information with regard to
        the component.
    position: int
        The location inside the level that the component is housed.
    """

    domain: int
    index: int
    position: PositionProtocol


def domain_check(instance, attribute, value):
    """
    A check to ensure that the domain of a given object is between 0 and 7.

    Raises
    ------
    TypeError
        Will be raised if the domain is not an integer.
    ValueError
        Will be raised if the domain is not between 0 and 7.
    """
    if not isinstance(value, int):
        raise TypeError(f"{instance.__class__.__name__} only supports integers inside the domain, not {type(value)}")
    if 0 > value < 7:
        raise ValueError(f"{instance.__class__.__name__} only supports domains from 0 to 7, not {value}")


def index_check(instance, attribute, value):
    """
    A check to ensure that the index of a given object is between 0 and 255.

    Raises
    ------
    TypeError
        Will be raised if the index is not an integer.
    ValueError
        Will be raised if the index is not between 0 and 255.
    """
    if not isinstance(value, int):
        raise TypeError(f"{instance.__class__.__name__} only supports integers inside the index, not {type(value)}")
    if 0 > value < 0xFF:
        raise ValueError(f"{instance.__class__.__name__} only supports index from 0 to 255, not {value}")


def position_check(instance, attribute, value):
    """
    A check to ensure that the position of a given object has an x and y
    that are between 0 and 255.

    Raises
    ------
    ValueError
        Will be raised if either the x or y is not between 0 and 255.
    AttributeError
        Will be raised if the position does not contain a field for x or y.
    """
    try:
        if 0 > value.x < 0xFF:
            raise ValueError(f"{instance.__class__.__name__} only supports x from 0 to 255, not {value.x}")
    except AttributeError:
        raise AttributeError(f"{value.__class__.__name__} does not contain the field x")
    try:
        if 0 > value.y < 27:
            raise ValueError(f"{instance.__class__.__name__} only supports y from 0 to 27, not {value.y}")
    except AttributeError:
        raise AttributeError(f"{value.__class__.__name__} does not contain the field y")


@attr.s(slots=True)
class LevelComponent:
    """
    A basic implementation of `~foundry.smb3parse.objects.LevelComponentProtocol`.

    Attributes
    ----------
    domain: int
        The domain or set of component that the component is housed inside.
        Must be between 0 and 7.
    index: int
        The specific byte of data that determines a limited amount of information with regard to
        the component.
        Must be between 0 and 255.
    position: int
        The location inside the level that the component is housed.
        Both x and y must be between 0 and 255.
    """

    domain: int = attr.ib(validator=domain_check, default=0)
    index: int = attr.ib(validator=index_check, default=0)
    position: Position = attr.ib(validator=position_check, default=attr.Factory(lambda: Position(0, 0)))


class InLevelObject(ABC):
    """
    Describes objects that are positioned at a specific place in a level and have some sort of representation, be it
    visible like platforms and enemies or invisible, like auto scroll items or jumps.
    """

    def __init__(self, data: bytearray):
        self._data: bytearray = data
        self.level_component = LevelComponent()
        self._length: Optional[int] = None

    @property
    def id(self):
        return self.level_component.index

    @id.setter
    def id(self, value):
        self.level_component.index = value

    @property
    def domain(self):
        return self.level_component.domain

    @domain.setter
    def domain(self, value):
        self.level_component.domain = value

    @property
    def x(self):
        return self.level_component.position.x

    @x.setter
    def x(self, value):
        self.level_component.position.x = value

    @property
    def y(self):
        return self.level_component.position.y

    @y.setter
    def y(self, value):
        self.level_component.position.y = value

    @property
    def additional_length(self):
        return self._length

    @additional_length.setter
    def additional_length(self, value):
        self._length = value

    @property
    def has_additional_length(self):
        return self.additional_length is not None



from typing import Optional

from PySide6.QtCore import Signal, SignalInstance
from PySide6.QtWidgets import QHBoxLayout, QWidget

from foundry.game.gfx.Palette import PaletteProtocol
from foundry.gui.ColorButtonWidget import ColorButtonWidget


class PaletteWidget(QWidget):
    """
    A widget to view the palette.

    Attributes
    ----------
    palette_changed: Signal[Palette]
        Slot associated with the palette viewer changing its palette.
    """

    palette_changed: SignalInstance = Signal(PaletteProtocol)  # type: ignore

    def __init__(self, parent: Optional[QWidget], palette: PaletteProtocol):
        super().__init__(parent)
        self._palette = palette

        layout = QHBoxLayout(self)
        layout.setContentsMargins(1, 2, 0, 2)

        self._buttons = [ColorButtonWidget(self, color) for color in palette.colors]

        for button in self._buttons:
            layout.addWidget(button)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.parent}, {self.palette})"

    @property
    def palette(self) -> PaletteProtocol:
        return self._palette

    @palette.setter
    def palette(self, palette: PaletteProtocol):
        self._palette = palette
        self.palette_changed.emit(palette)
        self._update()

    def _update(self):
        for idx, color in enumerate(self.palette.colors):
            self._buttons[idx].color = color



from typing import Tuple


class WorldMapPosition:
    def __init__(self, world, screen: int, row: int, column: int):
        self.world = world
        self.screen = screen
        self.row = row
        self.column = column

    @property
    def level_info(self) -> Tuple[int, int, int]:
        return self.world.level_for_position(self.screen, self.row, self.column)

    def can_have_level(self):
        return self.world.is_enterable(self.tile())

    def tile(self):
        return self.world.tile_at(self.screen, self.row, self.column)

    def tuple(self):
        return self.world.number, self.screen, self.row, self.column

    def __eq__(self, other):
        if not isinstance(other, WorldMapPosition):
            return False

        return (
            self.world.number == other.world.number
            and self.screen == other.screen
            and self.row == other.row
            and self.column == other.column
        )

    def __repr__(self):
        return f"WorldMapPosition({self.world}, screen={self.screen}, row={self.row}, column={self.column})"



from typing import Protocol

from attr import attrs
from PySide6.QtGui import QColor, QImage, QPainter

from foundry.core.Position import PositionProtocol
from foundry.core.Size import Size, SizeProtocol
from foundry.core.sprites import SPRITE_SIZE
from foundry.core.sprites.Sprite import SpriteProtocol
from foundry.game.gfx.drawable import MASK_COLOR
from foundry.game.gfx.drawable.Sprite import Sprite as MetaSprite
from foundry.game.gfx.GraphicsSet import GraphicsSetProtocol
from foundry.game.gfx.Palette import PaletteGroupProtocol


class SpriteGroupProtocol(Protocol):
    position: PositionProtocol
    sprites: list[SpriteProtocol]
    graphics_set: GraphicsSetProtocol
    palette_group: PaletteGroupProtocol

    @property
    def size(self) -> SizeProtocol:
        ...

    def image(self, scale_factor: int) -> QImage:
        ...


@attrs(slots=True, auto_attribs=True)
class SpriteGroup:
    """
    A representation of a group of sprites inside the game.

    Attributes
    ----------
    position: PositionProtocol
        The position of the sprite group.
    sprites: list[SpriteProtocol]
        The sprites that compose the sprite group.
    graphics_set: GraphicsSetProtocol
        The graphics to render the sprites with.
    palette_group: PaletteGroupProtocol
        The palettes to render the sprites with.
    """

    position: PositionProtocol
    sprites: list[SpriteProtocol]
    graphics_set: GraphicsSetProtocol
    palette_group: PaletteGroupProtocol

    @property
    def size(self) -> SizeProtocol:
        return Size(
            max([sprites.position.x for sprites in self.sprites]) + SPRITE_SIZE.width,
            max([sprites.position.y for sprites in self.sprites]) + SPRITE_SIZE.height,
        )

    def image(self, scale_factor: int = 1) -> QImage:
        image = QImage(self.size.width * scale_factor, self.size.height * scale_factor, QImage.Format_RGB888)
        image.fill(QColor(*MASK_COLOR))
        painter = QPainter(image)

        for sprite_data in self.sprites:
            if sprite_data.do_not_render:
                continue
            else:
                sprite = MetaSprite(
                    sprite_data.index,
                    tuple(tuple(c for c in pal) for pal in self.palette_group),  # type: ignore
                    sprite_data.palette_index,
                    self.graphics_set,
                    sprite_data.horizontal_mirror,
                    sprite_data.vertical_mirror,
                )
                sprite.draw(
                    painter,
                    sprite_data.position.x * scale_factor,
                    sprite_data.position.y * scale_factor,
                    SPRITE_SIZE.width * scale_factor,
                    SPRITE_SIZE.height * scale_factor,
                    transparent=True,
                )

        painter.end()
        return image



from PySide6.QtCore import QPoint, QPointF, QRect
from PySide6.QtGui import QColor, QPainter, QPen, Qt

STROKE_COLOR = QColor(0x00, 0x00, 0x00, 0x80)


class SelectionSquare:
    def __init__(self):
        self.start_point = QPoint(0, 0)
        self.end_point = QPoint(0, 0)

        self.active = False
        self.should_draw = False

        self.rect = QRect(self.start_point, self.end_point)

        self.pen = QPen(STROKE_COLOR)
        self.pen.setWidth(1)
        self.brush = Qt.NoBrush

    def is_active(self):
        return self.active

    def start(self, point: QPoint):
        self.active = True

        # Convert to a regular point
        if isinstance(point, QPointF):
            point = point.toPoint()

        self.start_point = point

    def set_current_end(self, point: QPoint):
        if not self.active:
            return

        # Convert to a regular point
        if isinstance(point, QPointF):
            point = point.toPoint()

        self.should_draw = True

        self.end_point = point

        self.rect = QRect(self.start_point, self.end_point)

    def stop(self):
        self.active = False
        self.should_draw = False

    def get_rect(self):
        return self.rect

    def get_adjusted_rect(self, horizontal_factor: int, vertical_factor: int) -> QRect:
        x, y = self.get_rect().topLeft().toTuple()
        width, height = self.get_rect().size().toTuple()

        x //= horizontal_factor
        width //= horizontal_factor

        y //= vertical_factor
        height //= vertical_factor

        return QRect(x, y, width + 1, height + 1)

    def draw(self, painter: QPainter):
        if self.should_draw:
            painter.setPen(self.pen)
            painter.setBrush(self.brush)

            painter.drawRect(self.rect)



from typing import Optional

from foundry.smb3parse.levels.world_map import get_all_world_maps
from foundry.smb3parse.levels.WorldMapPosition import WorldMapPosition
from foundry.smb3parse.util.rom import Rom


def find_world_position(rom: Rom, layout_address: int) -> Optional[WorldMapPosition]:
    for world in get_all_world_maps(rom):
        for _level in world.gen_levels():
            if _level.layout_address == layout_address:
                return _level.world_map_position
    else:
        return None



import base64
import json
import os
import pathlib
import shlex
import subprocess
import tempfile
from typing import Optional

from PySide6.QtCore import QSize
from PySide6.QtGui import QAction, QCloseEvent, QKeySequence, QMouseEvent, QShortcut, Qt
from PySide6.QtWidgets import (
    QFileDialog,
    QMainWindow,
    QMenu,
    QMessageBox,
    QPushButton,
    QToolBar,
    QWhatsThis,
)

from foundry import (
    auto_save_level_data_path,
    auto_save_m3l_path,
    auto_save_rom_path,
    get_current_version_name,
    get_latest_version_name,
    icon,
    main_window_flags_path,
    open_url,
    releases_link,
)
from foundry.game.File import ROM
from foundry.game.level.LevelManager import LevelManager
from foundry.gui.AboutWindow import AboutDialog
from foundry.gui.ContextMenu import CMAction
from foundry.gui.LevelSelector import LevelSelector
from foundry.gui.settings import GUI_STYLE, SETTINGS, save_settings
from foundry.gui.SettingsDialog import POWERUPS, SettingsDialog
from foundry.gui.util import setup_window
from foundry.smb3parse.constants import (
    TILE_LEVEL_1,
    Title_DebugMenu,
    Title_PrepForWorldMap,
)
from foundry.smb3parse.levels.world_map import WorldMap as SMB3World
from foundry.smb3parse.util.rom import Rom as SMB3Rom

ROM_FILE_FILTER = "ROM files (*.nes *.rom);;All files (*)"
M3L_FILE_FILTER = "M3L files (*.m3l);;All files (*)"
IMG_FILE_FILTER = "Screenshots (*.png);;All files (*)"

with open(main_window_flags_path, "r") as data:
    main_window_flags = json.loads(data.read())
del data

menu_options_groups = main_window_flags["menu"]["view"]["options"]
menu_options = {}
for menu_option in menu_options_groups:
    menu_options.update(menu_option)

CHECKABLE_MENU_ITEMS = [option["id"] for option in menu_options.values()]
CHECKABLE_MENU = {option["id"]: option for option in menu_options.values()}
del menu_options_groups
del menu_options

ID_PROP = "ID"

# mouse modes

MODE_FREE = 0
MODE_DRAG = 1
MODE_RESIZE = 2


class MainWindow(QMainWindow):
    level_selector_last_level: Optional[tuple[int, int]]

    def __init__(self, path_to_rom="", world=None, level=None):
        super(MainWindow, self).__init__()

        self.setWindowIcon(icon(main_window_flags["icon"]))
        try:
            GUI_STYLE[SETTINGS["gui_style"]](self)
        except KeyError:
            SETTINGS["gui_style"] = "LIGHT BLUE"
            GUI_STYLE[SETTINGS["gui_style"]]

        setup_window(self, main_window_flags)

        self.manager = LevelManager(self)
        self.manager.on_enable()

        self.menu_toolbar = QToolBar("Menu Toolbar", self)
        self.menu_toolbar.setOrientation(Qt.Horizontal)
        self.menu_toolbar.setIconSize(QSize(20, 20))

        self.menu_toolbar.addAction(icon("settings.svg"), "Editor Settings").triggered.connect(self._on_show_settings)
        self.menu_toolbar.addSeparator()
        self.menu_toolbar.addAction(icon("folder.svg"), "Open ROM").triggered.connect(self.on_open_rom)
        self.menu_toolbar_save_action = self.menu_toolbar.addAction(icon("save.svg"), "Save Level")
        self.menu_toolbar_save_action.triggered.connect(self.on_save_rom)
        self.menu_toolbar.addSeparator()

        self.undo_action = self.menu_toolbar.addAction(icon("rotate-ccw.svg"), "Undo Action")
        self.undo_action.triggered.connect(self.manager.undo)
        self.undo_action.setEnabled(False)
        self.redo_action = self.menu_toolbar.addAction(icon("rotate-cw.svg"), "Redo Action")
        self.redo_action.triggered.connect(self.manager.redo)
        self.redo_action.setEnabled(False)

        self.menu_toolbar.addSeparator()
        play_action = self.menu_toolbar.addAction(icon("play-circle.svg"), "Play Level")
        play_action.triggered.connect(self.on_play)
        play_action.setWhatsThis("Opens an emulator with the current Level set to 1-1.\nSee Settings.")
        self.menu_toolbar.addSeparator()
        self.menu_toolbar.addAction(icon("zoom-out.svg"), "Zoom Out").triggered.connect(self.manager.zoom_out)
        self.menu_toolbar.addAction(icon("zoom-in.svg"), "Zoom In").triggered.connect(self.manager.zoom_in)
        self.menu_toolbar.addSeparator()
        header_action = self.menu_toolbar.addAction(icon("tool.svg"), "Edit Level Header")
        header_action.triggered.connect(self.manager.display_header_editor)
        header_action.setWhatsThis(
            "<b>Header Editor</b><br/>"
            "Many configurations regarding the level are done in its header, like the length of "
            "the timer, or where and how Mario enters the level.<br/>"
        )

        self.jump_destination_action = self.menu_toolbar.addAction(
            icon("arrow-right-circle.svg"), "Go to Jump Destination"
        )
        self.jump_destination_action.triggered.connect(self.manager.warp_to_alternative)
        self.jump_destination_action.setWhatsThis(
            "Opens the level, that can be reached from this one, e.g. by entering a pipe."
        )

        self.menu_toolbar.addSeparator()

        whats_this_action = QWhatsThis.createAction()
        whats_this_action.setWhatsThis("Click on parts of the editor, to receive help information.")
        whats_this_action.setIcon(icon("help-circle.svg"))
        whats_this_action.setText("Starts 'What's this?' mode")
        self.menu_toolbar.addAction(whats_this_action)

        self.warning_action = self.menu_toolbar.addAction(icon("alert-triangle.svg"), "Warning Panel")
        self.warning_action.setWhatsThis("Shows a list of warnings.")
        self.warning_action.setDisabled(True)
        self.warning_action.triggered.connect(self.manager.display_warnings)

        self.addToolBar(Qt.TopToolBarArea, self.menu_toolbar)

        self.delete_shortcut = QShortcut(QKeySequence(Qt.Key_Delete), self, self.manager.delete)

        QShortcut(QKeySequence(Qt.CTRL + Qt.Key_X), self, self.manager.cut)
        QShortcut(QKeySequence(Qt.CTRL + Qt.Key_C), self, self.manager.copy)
        QShortcut(QKeySequence(Qt.CTRL + Qt.Key_V), self, self.manager.paste)

        QShortcut(QKeySequence(Qt.CTRL + Qt.Key_Z), self, self.manager.undo)
        QShortcut(QKeySequence(Qt.CTRL + Qt.Key_Y), self, self.manager.redo)
        QShortcut(QKeySequence(Qt.CTRL + Qt.SHIFT + Qt.Key_Z), self, self.manager.redo)

        QShortcut(QKeySequence(Qt.CTRL + Qt.Key_Plus), self, self.manager.zoom_in)
        QShortcut(QKeySequence(Qt.CTRL + Qt.Key_Minus), self, self.manager.zoom_out)

        QShortcut(QKeySequence(Qt.CTRL + Qt.Key_A), self, self.manager.select_all)
        QShortcut(QKeySequence(Qt.CTRL + Qt.Key_L), self, self.manager.focus_selected)

        self.loaded = self.on_open_rom(path_to_rom, world, level)

        self.showMaximized()

    def _on_show_settings(self):
        SettingsDialog(self).exec_()

    @staticmethod
    def _save_auto_rom():
        ROM().save_to_file(auto_save_rom_path, set_new_path=False)

    def _save_auto_data(self):
        if self.manager.controller is None:
            return

        (object_offset, object_data), (enemy_offset, enemy_data) = self.manager.controller.level_ref.level.to_bytes()

        object_set_number = self.manager.controller.level_ref.level.object_set_number

        base64_data = []
        base64_data.append(
            (
                object_offset,
                base64.b64encode(object_data).decode("ascii"),
                enemy_offset,
                base64.b64encode(enemy_data).decode("ascii"),
            )
        )

        with open(auto_save_level_data_path, "w") as level_data_file:
            level_data_file.write(json.dumps([object_set_number, object_offset, enemy_offset, base64_data]))

    def _load_auto_save(self):
        # rom already loaded
        with open(auto_save_level_data_path, "r") as level_data_file:
            json_data = level_data_file.read()

            object_set_number, level_offset, enemy_offset, base64_data = json.loads(json_data)

        # load level from ROM, or from m3l file
        if level_offset == enemy_offset == 0:
            if not auto_save_m3l_path.exists():
                QMessageBox.critical(
                    self,
                    "Failed loading auto save",
                    "Could not recover m3l file, that was edited, when the editor crashed.",
                )

            with open(auto_save_m3l_path, "rb") as m3l_file:
                self.load_m3l(bytearray(m3l_file.read()), auto_save_m3l_path)
        else:
            self.update_level("recovered level", level_offset, enemy_offset, object_set_number)

        # restore undo/redo stack
        byte_data = []
        level_offset, object_data, enemy_offset, enemy_data = base64_data
        object_data = bytearray(base64.b64decode(object_data))
        enemy_data = bytearray(base64.b64decode(enemy_data))
        byte_data.append(((level_offset, object_data), (enemy_offset, enemy_data)))

        if self.manager.controller is None:
            return
        self.manager.controller.level_ref.changed = bool(base64_data)
        self.manager.controller.level_ref.do(((level_offset, object_data), (enemy_offset, enemy_data)))

    def on_play(self):
        """
        Copies the ROM, including the current level, to a temporary directory, saves the current level as level 1-1 and
        opens the rom in an emulator.
        """
        temp_dir = pathlib.Path(tempfile.gettempdir()) / "smb3foundry"
        temp_dir.mkdir(parents=True, exist_ok=True)

        path_to_temp_rom = temp_dir / "instaplay.rom"

        ROM().save_to(str(path_to_temp_rom))

        temp_rom = self._open_rom(path_to_temp_rom)

        self._put_current_level_to_level_1_1(temp_rom)
        self._set_default_powerup(temp_rom)

        self.side_palette.save(temp_rom)

        temp_rom.write(
            Title_PrepForWorldMap - 0x8,
            bytes([SETTINGS["default_starting_world"] - 1]),
        )

        temp_rom.save_to(str(path_to_temp_rom))

        arguments = SETTINGS["instaplay_arguments"].replace("%f", str(path_to_temp_rom))
        arguments = shlex.split(arguments, posix=False)

        emu_path = pathlib.Path(SETTINGS["instaplay_emulator"])

        if emu_path.is_absolute():
            if emu_path.exists():
                emulator = str(emu_path)
            else:
                QMessageBox.critical(
                    self, "Emulator not found", f"Check it under File > Settings.\nFile {emu_path} not found."
                )
                return
        else:
            emulator = SETTINGS["instaplay_emulator"]

        try:
            subprocess.run([emulator, *arguments])
        except Exception as e:
            QMessageBox.critical(self, "Emulator command failed.", f"Check it under File > Settings.\n{str(e)}")

    @staticmethod
    def _open_rom(path_to_rom):
        with open(path_to_rom, "rb") as smb3_rom:
            data = smb3_rom.read()

        rom = SMB3Rom(bytearray(data))
        return rom

    def _put_current_level_to_level_1_1(self, rom) -> bool:
        # load world data
        world = SMB3World.from_world_number(rom, SETTINGS["default_starting_world"])

        # find position of "level 1" tile in world map
        for position in world.gen_positions():
            if position.tile() == TILE_LEVEL_1:
                break
        else:
            QMessageBox.critical(
                self, "Couldn't place level", "Could not find a level 1 tile in World 1 to put your level at."
            )
            return False

        if self.manager.controller is not None and not self.manager.controller.level_ref.level.attached_to_rom:
            QMessageBox.critical(
                self,
                "Couldn't place level",
                "The Level is not part of the rom yet (M3L?). Try saving it into the ROM first.",
            )
            return False

        # write level and enemy data of current level
        if self.manager.controller is None:
            return False
        (layout_address, layout_bytes), (
            enemy_address,
            enemy_bytes,
        ) = self.manager.controller.level_ref.level.to_bytes()
        rom.write(layout_address, layout_bytes)
        rom.write(enemy_address, enemy_bytes)

        # replace level information with that of current level
        object_set_number = self.manager.controller.level_ref.level.object_set_number

        world.replace_level_at_position((layout_address, enemy_address - 1, object_set_number), position)

    def _set_default_powerup(self, rom) -> bool:
        *_, powerup, hasPWing = POWERUPS[SETTINGS["default_powerup"]]
        hasStar = SETTINGS["default_power_has_star"]
        nop = 0xEA

        rom.write(Title_PrepForWorldMap - 0x4, bytes([nop, nop, nop]))

        rom.write(Title_PrepForWorldMap + 0x1, bytes([powerup]))

        rts = 0x60
        lda = 0xA9
        staAbsolute = 0x8D
        Map_Power_DispHigh = 0x03
        Map_Power_DispLow = 0xF3

        # If a P-wing powerup is selected, another variable needs to be set with the P-wing value
        # This piece of code overwrites a part of Title_DebugMenu
        if hasPWing:
            # We need to start one byte before Title_DebugMenu to remove the RTS of Title_PrepForWorldMap
            # The assembly code below reads as follows:
            # LDA 0x08
            # STA $03F3
            # RTS
            if hasStar:
                rom.write(
                    Title_DebugMenu - 0x1,
                    bytes(
                        [
                            lda,
                            0x08,
                            staAbsolute,
                            Map_Power_DispLow,
                            Map_Power_DispHigh,
                            lda,
                            0x10,
                            staAbsolute,
                            Map_Power_DispLow - 1,
                            Map_Power_DispHigh,
                            rts,
                        ]
                    ),
                )
            else:
                rom.write(
                    Title_DebugMenu - 0x1, bytes([lda, 0x08, staAbsolute, Map_Power_DispLow, Map_Power_DispHigh, rts])
                )

        elif hasStar:
            rom.write(
                Title_DebugMenu - 0x1, bytes([lda, 0x10, staAbsolute, Map_Power_DispLow - 1, Map_Power_DispHigh, rts])
            )

        # Remove code that resets the powerup value by replacing it with no-operations
        # Otherwise this code would copy the value of the normal powerup here
        # (So if the powerup would be Raccoon Mario, Map_Power_Disp would also be
        # set as Raccoon Mario instead of P-wing
        Map_Power_DispResetLocation = 0x3C5A2
        rom.write(Map_Power_DispResetLocation, bytes([nop, nop, nop]))

    def on_screenshot(self, _) -> bool:
        recommended_file = f"{os.path.expanduser('~')}/{ROM.name} - {self.manager.title_suggestion}.png"

        pathname, _ = QFileDialog.getSaveFileName(
            self, caption="Save Screenshot", dir=recommended_file, filter=IMG_FILE_FILTER
        )

        if not pathname:
            return False

        # Proceed loading the file chosen by the user
        self.manager.screenshot.save(pathname)

        return True

    def update_title(self):
        self.setWindowTitle(self.manager.title_suggestion if self.manager.enabled else "SMB3Foundry")

    def on_open_rom(self, path_to_rom="", world: Optional[int] = None, level: Optional[int] = None) -> bool:
        if not self.safe_to_change():
            return False

        if not path_to_rom:
            # otherwise ask the user what new file to open
            path_to_rom, _ = QFileDialog.getOpenFileName(self, caption="Open ROM", filter=ROM_FILE_FILTER)
        if not path_to_rom:
            return False

        # Proceed loading the file chosen by the user
        try:
            ROM.load_from_file(path_to_rom)

            if path_to_rom == auto_save_rom_path:
                self._load_auto_save()
            else:
                self._save_auto_rom()
                if world is not None and level is not None:
                    self.manager.force_select(world, level)
                else:
                    self.manager.on_select()
            return True
        except IOError as exp:
            QMessageBox.warning(self, type(exp).__name__, f"Cannot open file '{path_to_rom}'.")
            return False

    def on_open_m3l(self, _) -> bool:
        if not self.safe_to_change():
            return False

        # otherwise ask the user what new file to open
        pathname, _ = QFileDialog.getOpenFileName(self, caption="Open M3L file", filter=M3L_FILE_FILTER)

        if not pathname:
            return False

        # Proceed loading the file chosen by the user
        try:
            with open(pathname, "rb") as m3l_file:

                m3l_data = bytearray(m3l_file.read())
        except IOError as exp:
            QMessageBox.warning(self, type(exp).__name__, f"Cannot open file '{pathname}'.")

            return False

        self.manager.load_m3l(m3l_data, pathname)
        self.save_m3l(auto_save_m3l_path, self.manager.to_m3l())

        return True

    def safe_to_change(self) -> bool:
        if self.manager.safe_to_change:
            return True
        else:
            answer = QMessageBox.question(
                self,
                "Please confirm",
                "Current content has not been saved! Proceed?",
                QMessageBox.No | QMessageBox.Yes,
                QMessageBox.No,
            )

            return answer == QMessageBox.Yes

    def _ask_for_palette_save(self) -> bool:
        """
        If Object Palettes have been changed, this function opens a dialog box, asking the user, if they want to save
        the changes, dismiss them, or cancel whatever they have been doing (probably saving/selecting another level).
        Saving or restoring Object Palettes is done inside the function if necessary.
        :return: False, if Cancel was chosen. True, if Palettes were restored or saved to ROM.
        """
        if hasattr(self, "side_palette") and not self.side_palette.changed:
            return True

        answer = QMessageBox.question(
            self,
            "Please confirm",
            "You changed some object palettes. This is a change, that potentially affects other levels in this ROM. Do "
            "you want to save these changes?",
            QMessageBox.Cancel | QMessageBox.RestoreDefaults | QMessageBox.Yes,
            QMessageBox.Cancel,
        )

        if answer == QMessageBox.Cancel:
            self.side_palette.restore()
            self.manager.refresh()
            return False

        if answer == QMessageBox.Yes:
            self.side_palette.save()
        elif answer == QMessageBox.RestoreDefaults:
            self.side_palette.restore()
            self.manager.refresh()

        return True

    def on_save_rom(self, _):
        self.save_rom(False)

    def on_save_rom_as(self, _):
        self.save_rom(True)

    def save_rom(self, is_save_as):
        safe_to_save, reason, additional_info = self.manager.stable_changes

        if not safe_to_save:
            answer = QMessageBox.warning(
                self,
                reason,
                f"{additional_info}\n\nDo you want to proceed?",
                QMessageBox.No | QMessageBox.Yes,
                QMessageBox.No,
            )

            if answer == QMessageBox.No:
                return
        self._ask_for_palette_save()

        if not self.manager.is_attached:
            QMessageBox.information(
                self,
                "Importing M3L into ROM",
                "You are currently editing a level stored in an m3l file outside of the ROM. Please select the "
                "positions in the ROM you want the level objects and enemies/items to be stored.",
                QMessageBox.Ok,
            )

            level_selector = LevelSelector(self)

            answer = level_selector.exec_()

            if answer == QMessageBox.Accepted:
                self.manager.attach(level_selector.object_data_offset, level_selector.enemy_data_offset)

                if is_save_as:
                    # if we save to another rom, don't consider the level
                    # attached (to the current rom)
                    self.manager.is_attached = False
                else:
                    # the m3l is saved to the current ROM, we can get rid of the auto save
                    auto_save_m3l_path.unlink(missing_ok=True)
            else:
                return

        if is_save_as:
            pathname, _ = QFileDialog.getSaveFileName(self, caption="Save ROM as", filter=ROM_FILE_FILTER)
            if not pathname:
                return  # the user changed their mind
        else:
            pathname = ROM.path

        if str(pathname) == str(auto_save_rom_path):
            QMessageBox.critical(
                self,
                "Cannot save to auto save ROM",
                "You can't save to the auto save ROM, as it will be deleted, when exiting the editor. Please choose "
                "another location, or your changes will be lost.",
            )

        self._save_current_changes_to_file(pathname, set_new_path=True)

        self.update_title()

        if not is_save_as:
            self.manager.safe_to_change = True

    def _save_current_changes_to_file(self, pathname: str, set_new_path):
        for data in self.manager.to_data():
            ROM().bulk_write(bytearray(data.data), data.location)

        try:
            ROM().save_to_file(pathname, set_new_path)

            self._save_auto_rom()
        except IOError as exp:
            QMessageBox.warning(self, f"{type(exp).__name__}", f"Cannot save ROM data to file '{pathname}'.")

    def on_save_m3l(self, _):
        suggested_file = self.manager.title_suggestion

        if not suggested_file.endswith(".m3l"):
            suggested_file += ".m3l"

        pathname, _ = QFileDialog.getSaveFileName(
            self, caption="Save M3L as", dir=suggested_file, filter=M3L_FILE_FILTER
        )

        if not pathname:
            return

        m3l_bytes = self.manager.to_m3l()

        self.save_m3l(pathname, m3l_bytes)

    def save_m3l(self, pathname: str, m3l_bytes: bytearray):
        try:
            with open(pathname, "wb") as m3l_file:
                m3l_file.write(m3l_bytes)
        except IOError as exp:
            QMessageBox.warning(self, type(exp).__name__, f"Couldn't save level to '{pathname}'.")

    def on_check_for_update(self):
        self.setCursor(Qt.WaitCursor)

        current_version = get_current_version_name()

        try:
            latest_version = get_latest_version_name()
        except ValueError as ve:
            QMessageBox.critical(self, "Error while checking for updates", f"Error: {ve}")
            self.setCursor(Qt.ArrowCursor)
            return

        if current_version != latest_version:
            latest_release_url = f"{releases_link}/tag/{latest_version}"

            go_to_github_button = QPushButton(icon("external-link.svg"), "Go to latest release")
            go_to_github_button.clicked.connect(lambda: open_url(latest_release_url))

            info_box = QMessageBox(
                QMessageBox.Information, "New release available", f"New Version {latest_version} is available."
            )

            info_box.addButton(QMessageBox.Cancel)
            info_box.addButton(go_to_github_button, QMessageBox.AcceptRole)

            info_box.exec_()
        else:
            QMessageBox.information(self, "No newer release", f"Version {current_version} is up to date.")

        self.setCursor(Qt.ArrowCursor)

    def reload_level(self):
        self.manager.refresh()

    def on_menu(self, action: QAction):
        item_id = action.property(ID_PROP)

        if item_id in CHECKABLE_MENU_ITEMS:
            self.on_menu_item_checked(action)

            # if setting a checkbox, keep the menu open
            menu_of_action: QMenu = self.sender()
            menu_of_action.exec_()

        elif item_id in self.manager.actions:
            x, y = self.manager.last_position

            if item_id == CMAction.REMOVE:
                self.manager.delete()
            elif item_id == CMAction.ADD_OBJECT:
                self.manager.create_object_from_suggestion((x, y))
            elif item_id == CMAction.CUT:
                self.manager.cut()
            elif item_id == CMAction.COPY:
                self.manager.copy()
            elif item_id == CMAction.PASTE:
                self.manager.paste(x, y)
            elif item_id == CMAction.FOREGROUND:
                self.manager.to_foreground()
            elif item_id == CMAction.BACKGROUND:
                self.manager.to_background()

        self.manager.update()

    def on_menu_item_checked(self, action: QAction):
        item_id = action.property(ID_PROP)

        checked = action.isChecked()

        if item_id not in CHECKABLE_MENU:
            return
        setattr(self.level_view, CHECKABLE_MENU[item_id]["attribute"], checked)
        SETTINGS[CHECKABLE_MENU[item_id]["name"]] = checked

        save_settings()

    def open_level_selector(self, _):
        self.manager.on_select()

    def on_player_viewer(self):
        self.manager.display_player_viewer()

    def on_object_viewer(self):
        self.manager.display_object_viewer()

    def on_block_viewer(self, _):
        self.manager.display_block_viewer()

    def on_palette_viewer(self, _):
        self.manager.display_palette_viewer()

    def on_edit_autoscroll(self, _):
        self.manager.display_autoscroll_editor()

    def on_header_editor(self, _):
        self.manager.display_header_editor()

    def on_jump_edit(self):
        self.manager.display_jump_editor()

    def mouseReleaseEvent(self, event: QMouseEvent):
        if event.button() == Qt.MiddleButton:
            self.manager.middle_mouse_release(self.mapToGlobal(event.position()))

    def on_about(self, _):
        about = AboutDialog(self)
        about.show()

    def closeEvent(self, event: QCloseEvent):
        if not self.safe_to_change():
            event.ignore()

            return

        auto_save_rom_path.unlink(missing_ok=True)
        auto_save_m3l_path.unlink(missing_ok=True)
        auto_save_level_data_path.unlink(missing_ok=True)

        super(MainWindow, self).closeEvent(event)



from typing import Optional

from attr import attrs

from foundry.core.warnings.Warning import PydanticWarning, Warning
from foundry.game.gfx.objects.ObjectLike import ObjectLike


@attrs(slots=True, frozen=True, auto_attribs=True)
class InvalidPositionWarning(Warning):
    """
    A warning for an object exceeding its valid positional range.
    """

    max_x: Optional[int] = None
    min_x: Optional[int] = None
    max_y: Optional[int] = None
    min_y: Optional[int] = None

    def check_object(self, obj: ObjectLike, *args, **kwargs) -> bool:
        """
        Determines if the object should emit a warning for having an invalid position.

        Parameters
        ----------
        obj : ObjectLike
            The object to check.

        Returns
        -------
        bool
            If the object should emit a warning.
        """
        return (
            self.max_x is not None
            and obj.position.x > self.max_x
            or self.min_x is not None
            and obj.position.x < self.min_x
            or self.max_y is not None
            and obj.position.y > self.max_y
            or self.min_y is not None
            and obj.position.y < self.min_y
        )

    def get_message(self, obj: ObjectLike) -> str:
        if self.max_x is not None and obj.position.x > self.max_x:
            return f"{obj.name} x position of {obj.position.x} is more than its safe maximum of {self.max_x}."
        if self.min_x is not None and obj.position.x < self.min_x:
            return f"{obj.name} x position of {obj.position.x} is less than its safe minimum of {self.min_x}."
        if self.max_y is not None and obj.position.y > self.max_y:
            return f"{obj.name} y position of {obj.position.y} is more than its safe maximum of {self.max_y}."
        if self.min_y is not None and obj.position.y < self.min_y:
            return f"{obj.name} x position of {obj.position.y} is less than its safe maximum of {self.min_y}."
        raise NotImplementedError


class PydanticInvalidPositionWarning(PydanticWarning):
    """
    A JSON model of a warning that checks for an object being in a positional range.
    """

    max_x: Optional[int] = None
    min_x: Optional[int] = None
    max_y: Optional[int] = None
    min_y: Optional[int] = None



import json
import urllib.error
import urllib.request
from pathlib import Path
from typing import Union

from PySide6.QtCore import QUrl
from PySide6.QtGui import QDesktopServices, QIcon

root_dir = Path(__file__).parent

home_dir = Path.home() / ".smb3foundry"
home_dir.mkdir(parents=True, exist_ok=True)

default_settings_path = home_dir / "settings"

auto_save_path = home_dir / "auto_save"
auto_save_path.mkdir(parents=True, exist_ok=True)

auto_save_rom_path = auto_save_path / "auto_save.nes"
auto_save_m3l_path = auto_save_path / "auto_save.m3l"
auto_save_level_data_path = auto_save_path / "level_data.json"

data_dir = root_dir / "data"
main_window_flags_path = data_dir / "main_window_flags.json"
jump_creator_flags_path = data_dir / "jump_creator_flags.json"
spinner_panel_flags_path = data_dir / "spinner_panel_flags.json"
enemy_definitions = data_dir / "enemy_definitions.json"
warp_definitions = data_dir / "warp_definitions.json"
tileset_definitions = data_dir / "definitions.json"
tileset_data_path = data_dir / "tilesets.json"
doc_dir = root_dir.joinpath("doc")
icon_dir = data_dir.joinpath("icons")

releases_link = "https://github.com/TheJoeSmo/Foundry/releases"
feature_video_link = "https://www.youtube.com/watch?v=7_22cAffMmE"
github_link = "https://github.com/TheJoeSmo/Foundry"
github_issue_link = "https://github.com/TheJoeSmo/Foundry/issues"
discord_link = "https://discord.gg/pm87gm7"

enemy_compat_link = QUrl.fromLocalFile(str(doc_dir.joinpath("SMB3 enemy compatibility.html")))


def open_url(url: Union[str, QUrl]):
    QDesktopServices.openUrl(QUrl(url))


def get_current_version_name() -> str:
    from single_source import get_version

    return get_version("foundry_smb3", root_dir, fail=True)  # type: ignore


def get_latest_version_name(timeout: int = 10) -> str:
    owner = "TheJoeSmo"
    repo = "Foundry"

    api_call = f"https://api.github.com/repos/{owner}/{repo}/releases"

    try:
        request = urllib.request.urlopen(api_call, timeout=timeout)
    except urllib.error.URLError as ue:
        raise ValueError(f"Network error {ue}")

    data = request.read()

    try:
        return json.loads(data)[0]["tag_name"].strip()
    except (KeyError, IndexError, json.JSONDecodeError):
        raise ValueError("Parsing the received information failed.")


def icon(icon_name: str):
    icon_path = icon_dir / icon_name
    data_path = data_dir / icon_name

    if icon_path.exists():
        return QIcon(str(icon_path))
    elif data_path.exists():
        return QIcon(str(data_path))
    else:
        raise FileNotFoundError(icon_path)
